
from sqlalchemy import types, Column, ForeignKey, Integer, String, Float, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship

Base = declarative_base()


class RunMetadata(Base):
    __tablename__ = 'RunMetadata'

    run_id = Column(Integer, primary_key=True)
    run_name = Column(String(250), nullable=False, index=True)
    watershed_id = Column(Integer,
                          ForeignKey('Watershed.watershed_id'),
                          nullable=True)
    pixel = Column(Integer, nullable=True)
    description = Column(String(250), nullable=True)
    smrf_version = Column(String(250), nullable=True)
    awsm_version = Column(String(250), nullable=True)
    snowav_version = Column(String(250), nullable=True)
    data_type = Column(String(250), nullable=True)
    data_location = Column(String(1200), nullable=True)
    file_type = Column(String(250), nullable=True)
    config_file = Column(String(250), nullable=True)
    proc_time = Column(DateTime, nullable=True)

    # This puts Watershed fields available through Basin at Basin.watershed
    run_metadata = relationship('Watershed',
                                backref=backref('RunMetadata', lazy='dynamic'))


class Watershed(Base):
    __tablename__ = 'Watershed'

    watershed_id = Column(Integer, primary_key=True)
    watershed_name = Column(String(250), unique=True, nullable=False,
                            index=True)
    shapefile = Column(types.LargeBinary, nullable=True)


class Basin(Base):
    __tablename__ = 'Basin'

    watershed_id = Column(Integer,
                          ForeignKey('Watershed.watershed_id',
                                     onupdate="cascade"),
                          nullable=False)
    basin_id = Column(Integer, primary_key=True, nullable=False)
    basin_name = Column(String(250), nullable=False)
    shapefile = Column(types.LargeBinary, nullable=True)

    watershed = relationship('Watershed',
                             backref=backref('Basin', lazy='dynamic'))


class Results(Base):
    __tablename__ = 'Results'

    id = Column(Integer, primary_key=True)
    basin_id = Column(Integer,
                      ForeignKey('Basin.basin_id', onupdate="cascade"),
                      index=True)
    run_id = Column(Integer,
                    ForeignKey('RunMetadata.run_id', onupdate="cascade"),
                    index=True)
    date_time = Column(DateTime, nullable=False, index=True)
    variable = Column(String(250), nullable=True, index=True)
    variable_id = Column(Integer,
                         ForeignKey('VariableUnits.id', onupdate="cascade"),
                         nullable=True)
    value = Column(Float, nullable=True)
    elevation = Column(String(250), nullable=True)

    variable_units = relationship('VariableUnits',
                                  backref=backref('Results'))

    runmetadata = relationship('RunMetadata',
                               backref=backref('Results', lazy='dynamic'))

    basinid = relationship('Basin',
                           backref=backref('Results', lazy='dynamic'))


class VariableUnits(Base):
    __tablename__ = 'VariableUnits'

    id = Column(Integer, primary_key=True, autoincrement=True)
    run_id = Column(Integer,
                    ForeignKey('RunMetadata.run_id', onupdate="cascade"),
                    index=True)
    variable = Column(String(250), nullable=True)
    unit = Column(String(250), nullable=True)
    name = Column(String(250), nullable=True)

    run_metadata = relationship('RunMetadata',
                                backref=backref('VariableUnits', lazy='dynamic'))


class Inputs(Base):
    __tablename__ = 'Inputs'

    id = Column(Integer, primary_key=True)
    run_id = Column(Integer, nullable=False)
    run_name = Column(String(250), nullable=False, index=True)
    basin_id = Column(Integer, nullable=False, index=True)
    date_time = Column(DateTime, nullable=False, index=True)
    variable = Column(String(250), nullable=False)
    function = Column(String(250), nullable=False)
    value = Column(Float, nullable=True)
    unit = Column(String(250), nullable=True)


class Pixels(Base):
    __tablename__ = 'Pixels'

    id = Column(Integer, primary_key=True, unique=True)
    location = Column(String(250), nullable=False)
    model_row = Column(Integer, nullable=False)
    model_col = Column(Integer, nullable=False)
    description = Column(String(250), nullable=True)
    name = Column(String(250), nullable=True)
    utm_x = Column(Float, nullable=False)
    utm_y = Column(Float, nullable=False)
    elevation = Column(Float, nullable=False)


class PixelsData(Base):
    __tablename__ = 'PixelsData'

    pid = Column(Integer, primary_key=True)
    pixel_id = Column(Integer, ForeignKey('Pixels.id'), nullable=False)
    date_time = Column(DateTime, nullable=False, index=True)
    pixels = relationship(Pixels, foreign_keys=[pixel_id])

    # smrf
    air_temp = Column(Float, nullable=True)
    albedo_vis = Column(Float, nullable=True)
    albedo_ir = Column(Float, nullable=True)
    clear_ir_beam = Column(Float, nullable=True)
    clear_ir_diffuse = Column(Float, nullable=True)
    cloud_factor = Column(Float, nullable=True)
    cloud_ir_beam = Column(Float, nullable=True)
    cloud_ir_diffuse = Column(Float, nullable=True)
    cloud_vis_beam = Column(Float, nullable=True)
    cloud_vis_diffuse = Column(Float, nullable=True)
    dew_point = Column(Float, nullable=True)
    flatwind = Column(Float, nullable=True)
    net_solar = Column(Float, nullable=True)
    percent_snow = Column(Float, nullable=True)
    precip = Column(Float, nullable=True)
    precip_temp = Column(Float, nullable=True)
    snow_density = Column(Float, nullable=True)
    storm_days = Column(Float, nullable=True)
    thermal = Column(Float, nullable=True)
    thermal_clear = Column(Float, nullable=True)
    thermal_veg = Column(Float, nullable=True)
    thermal_cloud = Column(Float, nullable=True)
    vapor_pressure = Column(Float, nullable=True)
    veg_ir_diffuse = Column(Float, nullable=True)
    veg_ir_beam = Column(Float, nullable=True)
    veg_vis_beam = Column(Float, nullable=True)
    veg_vis_diffuse = Column(Float, nullable=True)
    wind_direction = Column(Float, nullable=True)
    wind_speed = Column(Float, nullable=True)

    # em.nc
    net_rad = Column(Float, nullable=True)
    sensible_heat = Column(Float, nullable=True)
    latent_heat = Column(Float, nullable=True)
    snow_soil = Column(Float, nullable=True)
    precip_advected = Column(Float, nullable=True)
    sum_EB = Column(Float, nullable=True)
    evaporation = Column(Float, nullable=True)
    snowmelt = Column(Float, nullable=True)
    SWI = Column(Float, nullable=True)
    cold_content = Column(Float, nullable=True)

    # snow.nc
    thickness = Column(Float, nullable=True)
    density = Column(Float, nullable=True)
    specific_mass = Column(Float, nullable=True)
    liquid_water = Column(Float, nullable=True)
    temp_surf = Column(Float, nullable=True)
    temp_lower = Column(Float, nullable=True)
    temp_snowcover = Column(Float, nullable=True)
    thickness_lower = Column(Float, nullable=True)
    water_saturation = Column(Float, nullable=True)



from copy import deepcopy
from datetime import datetime
import logging
import numpy as np
import pandas as pd
import os
from sys import exit
import warnings

from snowav.database.database import delete, query, package
from snowav.utils.utilities import calculate, sum_precip, snow_line, \
    input_summary
from tablizer.tablizer import get_existing_records


class Process(object):
    """ Process AWSM model results made in awsm_daily format.

    Args
    ------
    cfg: config object
    """

    def __init__(self, cfg):

        pd.options.mode.chained_assignment = None
        warnings.filterwarnings('ignore')

        elapsed_hours = 0
        variables = cfg.variables.variables

        if os.path.splitext(cfg.connector)[1] == '.db':
            db = 'sqlite'
        else:
            db = 'sql'

        # images for figures
        precip_total = np.zeros((cfg.nrows, cfg.ncols))
        rain_total = np.zeros((cfg.nrows, cfg.ncols))

        # Check that topo and outputs are the same dimensions
        if cfg.outputs['swe_z'][0].shape != cfg.dem.shape:
            logging.error(" topo.nc dimensions: {} do not match output "
                          "dimensions: "
                          "{}".format(cfg.dem.shape,
                                      cfg.outputs['swe_z'][0].shape))
            exit()

        # process each date
        for iters, out_date in enumerate(cfg.outputs['dates']):
            wy_hour = int(cfg.outputs['time'][iters])
            dir_str = cfg.rundirs_dict[wy_hour].split('/')[-1]
            odate_str = out_date.strftime("%Y-%-m-%-d %H:00")

            # assume to start that we are processing everything
            proc_list = deepcopy(cfg.variables.snowav_results_variables)
            pass_flag = False

            # delete anything on the database with the same run_name and date
            for bid in cfg.plotorder:
                results = query(cfg.connector, out_date, out_date,
                                cfg.run_name, cfg.basins, bid, 'swe_vol')

                if not results.empty:
                    if not cfg.db_overwrite:
                        # If database records exist we can skip all outputs
                        # but density on the last day, if making the figure
                        pass_flag = True

                        if bid == cfg.plotorder[0]:
                            logging.info(' Skipping outputs for {}, {}, '
                                         'database records exist'
                                         '...'.format(bid, odate_str))
                        else:
                            logging.debug(' Skipping outputs for {}, {}, '
                                          'database records exist'
                                          '...'.format(bid, odate_str))

                        if (out_date == cfg.outputs['dates'][-1] and
                                cfg.density_flag):
                            proc_list = ['density']
                        else:
                            proc_list = []

                    else:
                        out = delete(cfg.connector, cfg.basins, out_date,
                                     out_date, bid, cfg.run_name)
                        if len(out) > 0:
                            if bid == cfg.plotorder[0]:
                                logging.info(out[0])
                            else:
                                logging.debug(out[0])

            density = {}
            for name in cfg.masks:
                density[name] = {}

            # if database records don't exist, or we are making the
            # precip_depth figure we need to process precip
            if not pass_flag or cfg.precip_depth_flag:
                logging.info(' Processing precip {}'.format(
                    dir_str))
                run_dir = cfg.rundirs_dict[cfg.outputs['time'][iters]]
                precip_path = os.path.join(
                    run_dir.replace('/runs/run', '/data/data'),
                    'smrfOutputs/precip.nc')

                percent_snow_path = precip_path.replace('precip.nc',
                                                        'percent_snow.nc')

                if (os.path.isfile(precip_path) and
                        os.path.isfile(percent_snow_path)):
                    precip, rain = sum_precip(precip_path,
                                              percent_snow_path)
                    precip_total = precip_total + precip
                    rain_total = rain_total + rain
                else:
                    logging.warning(' One or both of precip.nc, '
                                    'percent_snow.nc does not exist')

            if cfg.inputs_flag:
                mask_list = []
                for name in cfg.masks:
                    mask_list.append(deepcopy(cfg.masks[name]['mask']))

                sf = cfg.rundirs_dict[wy_hour].replace('runs', 'data')
                sf = sf.replace('run', 'data') + '/smrfOutputs/'

                df = get_existing_records(cfg.connector, db)
                df = df.set_index('date_time')
                df.sort_index(inplace=True)

                for i, input in enumerate(cfg.variables.snowav_inputs_variables):

                    input_path = os.path.join(sf, input + '.nc')

                    for basin in cfg.inputs_basins:

                        basin_id = int(cfg.basins[basin]['basin_id'])

                        dfs = df[((df['variable'] == input) &
                                  (df['run_name'] == cfg.run_name) &
                                  (df['basin_id'] == basin_id))]
                        idx = dfs.index

                        if (not idx.contains(out_date) or
                                (idx.contains(out_date) and cfg.db_overwrite)):
                            if basin == cfg.inputs_basins[0] and i == 0:
                                logging.info(' Processing {} inputs, '
                                             '{} '.format(basin, odate_str))
                            else:
                                logging.debug(' Processing {}, {}, '
                                              '{}'.format(input, basin,
                                                          odate_str))

                            mask = cfg.masks[basin]['mask']

                            input_summary(input_path,
                                          input,
                                          cfg.inputs_methods,
                                          cfg.inputs_percentiles,
                                          db,
                                          cfg.connector,
                                          cfg.run_name,
                                          basin_id,
                                          cfg.run_id,
                                          mask)

                        if idx.contains(out_date) and not cfg.db_overwrite:
                            if ((basin == cfg.inputs_basins[0]) and
                                    (input == cfg.inputs_variables[0])):
                                logging.info(' Skipping inputs for {}, {}, {},'
                                             ' database records exist...'
                                             ''.format(input,
                                                       basin,
                                                       odate_str))
                            else:
                                logging.debug(' Skipping inputs for {}, {}, '
                                              '{}, database records exist...'
                                              ''.format(input,
                                                        basin,
                                                        odate_str))

            swe = deepcopy(cfg.outputs['swe_z'][iters])
            cold = deepcopy(cfg.outputs['coldcont'][iters])

            # Loop over outputs (depths are copied, volumes are calculated)
            for k in proc_list:

                # Mask by subbasin
                for name in cfg.masks:
                    if name == cfg.plotorder[0] and k == proc_list[0]:
                        logging.info(' Processing {}, {}'.format(name,
                                                                 dir_str))
                    else:
                        logging.debug(' Processing {}, {}, {}'.format(k, name,
                                                                      dir_str))

                    mask = deepcopy(cfg.masks[name]['mask'])
                    elevbin = cfg.ixd * mask

                    snow_mask = swe > 0
                    avail_mask = cold > cfg.cclimit
                    unavail_mask = cold <= cfg.cclimit

                    if k in cfg.variables.awsm_variables:
                        o = deepcopy(cfg.outputs[k][iters])

                    # elevation band
                    for n in np.arange(0, len(cfg.edges)):
                        b = cfg.edges[n]
                        elev_mask = elevbin == n

                        # list of masks
                        be = [mask, elev_mask]
                        ba = [mask, avail_mask]
                        bu = [mask, unavail_mask]
                        bes = [mask, elev_mask, snow_mask]
                        besa = [mask, elev_mask, snow_mask, avail_mask]
                        besu = [mask, elev_mask, snow_mask, unavail_mask]

                        if k == 'swe_z':
                            variables['swe_unavail']['df'].loc[b, name] = \
                                calculate(o, cfg.pixel, besu, 'sum', 'volume',
                                          cfg.units, cfg.dplcs)
                            variables['swe_avail']['df'].loc[b, name] = \
                                calculate(o, cfg.pixel, besa, 'sum', 'volume',
                                          cfg.units, cfg.dplcs)
                            variables['swe_vol']['df'].loc[b, name] = \
                                calculate(o, cfg.pixel, be, 'sum', 'volume',
                                          cfg.units, cfg.dplcs)
                            variables[k]['df'].loc[b, name] = \
                                calculate(o, cfg.pixel, be, 'mean', 'depth',
                                          cfg.units, cfg.dplcs)

                        if k == 'swi_z':
                            variables[k]['df'].loc[b, name] = \
                                calculate(o, cfg.pixel, be, 'mean', 'depth',
                                          cfg.units, cfg.dplcs)
                            variables['swi_vol']['df'].loc[b, name] = \
                                calculate(o, cfg.pixel, be, 'sum', 'volume',
                                          cfg.units, cfg.dplcs)

                        if k in cfg.variables.process_depth_units:

                            # iSnobal depth units are m
                            if k == 'depth':
                                type = 'snow_depth'
                            else:
                                type = variables[k]['unit_type']

                            variables[k]['df'].loc[b, name] = \
                                calculate(o, cfg.pixel, bes,
                                          variables[k]['calculate'], type,
                                          cfg.units, cfg.dplcs)

                            if (out_date == cfg.outputs['dates'][-1] and
                                    k == 'density'):
                                od = deepcopy(o)
                                ml = [mask, elev_mask, snow_mask]
                                for m in ml:
                                    m = m.astype('float')
                                    m[m < 1] = np.nan
                                    od = od * m

                                density[name][cfg.edges[n]] = deepcopy(od)

                        if k == 'precip_z':
                            variables['precip_z']['df'].loc[b, name] = \
                                calculate(precip, cfg.pixel, be,
                                          variables[k]['calculate'],
                                          variables[k]['unit_type'], cfg.units,
                                          cfg.dplcs)
                            variables['rain_z']['df'].loc[b, name] = \
                                calculate(rain, cfg.pixel, be,
                                          variables[k]['calculate'],
                                          variables[k]['unit_type'], cfg.units,
                                          cfg.dplcs)
                            variables['precip_vol']['df'].loc[b, name] = \
                                calculate(precip, cfg.pixel, be, 'sum',
                                          'volume', cfg.units, cfg.dplcs)

                    if k in cfg.variables.process_depth_units:

                        if k == 'depth':
                            type = 'snow_depth'
                        else:
                            type = variables[k]['unit_type']
                        variables[k]['df'].loc['total', name] = \
                            calculate(o, cfg.pixel, [mask, snow_mask],
                                      variables[k]['calculate'], type,
                                      cfg.units, cfg.dplcs)

                    if k == 'swe_z':
                        variables['swe_vol']['df'].loc['total', name] = \
                            calculate(o, cfg.pixel, mask, 'sum', 'volume',
                                      cfg.units, cfg.dplcs)
                        variables['swe_avail']['df'].loc['total', name] = \
                            calculate(o, cfg.pixel, ba, 'sum', 'volume',
                                      cfg.units, cfg.dplcs)
                        variables['swe_unavail']['df'].loc['total', name] = \
                            calculate(o, cfg.pixel, bu, 'sum', 'volume',
                                      cfg.units, cfg.dplcs)
                        variables[k]['df'].loc['total', name] = \
                            calculate(o, cfg.pixel, mask, 'mean', 'depth',
                                      cfg.units, cfg.dplcs)

                        if 'snow_line' in proc_list:
                            variables['snow_line']['df'].loc['total', name] = \
                                snow_line(o, cfg.dem, mask,
                                          cfg.diagnostics_flag)

                    if k == 'swi_z':
                        variables[k]['df'].loc['total', name] = \
                            calculate(o, cfg.pixel, mask, 'mean', 'depth',
                                      cfg.units, cfg.dplcs)
                        variables['swi_vol']['df'].loc['total', name] = \
                            calculate(o, cfg.pixel, mask, 'sum', 'volume',
                                      cfg.units, cfg.dplcs)

                    if k == 'precip_z':
                        variables['precip_z']['df'].loc['total', name] = \
                            calculate(precip, cfg.pixel, mask, 'mean', 'depth',
                                      cfg.units, cfg.dplcs)
                        variables['precip_vol']['df'].loc['total', name] = \
                            calculate(precip, cfg.pixel, mask, 'sum', 'volume',
                                      cfg.units, cfg.dplcs)
                        variables['rain_z']['df'].loc['total', name] = \
                            calculate(rain, cfg.pixel, mask, 'mean', 'depth',
                                      cfg.units, cfg.dplcs)

                df = deepcopy(variables[k]['df'])
                df = df.round(decimals=cfg.dplcs)

                if k == 'snow_line':
                    df = df[df.index == 'total']

                if not pass_flag:
                    package(cfg.connector, cfg.basins, df, cfg.run_id, cfg.vid,
                            k, out_date)

            stamp = datetime.now().strftime("%Y-%-m-%-d %H:%M:%S")
            if not pass_flag:
                logging.info(' Completed {} at {} UTC'.format(dir_str, stamp))

                if iters != 0 and int(wy_hour - elapsed_hours) != 24:
                    logging.warning(' Elapsed hours: {}, there may be a '
                                    'missing output '
                                    'day'.format(str(wy_hour - elapsed_hours)))

            elapsed_hours = int(wy_hour)

        self.density = density
        self.precip_total = precip_total
        self.rain_total = rain_total



import numpy as np

def nashsutcliffe(observed,modeled):
    """
    Nash-Sutcliffe model efficinecy

        .. math::

         NSE = 1-\\frac{\\sum_{i=1}^{N}(e_{i}-s_{i})^2}{\\sum_{i=1}^{N}(e_{i}-\\bar{e})^2}

    :observed: Observed data
    :type: list

    :simulation: Modeled data
    :type: list

    :return: Nash-Sutcliff model efficiency
    :rtype: float

    """
    if len(observed) == len(modeled):
        m = np.array(modeled)
        o = np.array(observed)

        mean_observed = np.nanmean(o)

        numerator = np.nansum((m - o) ** 2)
        denominator = np.nansum((o - mean_observed)**2)

        return 1 - (numerator / denominator)

    else:
        logging.warning("evaluation and simulation lists does not have the same length.")
        return np.nan



import pandas as pd
import numpy as np
from snowav.utils.wyhr import calculate_wyhr_from_date
import os
import glob
import pytz
from netCDF4 import Dataset
import netCDF4 as nc
import warnings

class iSnobalReader():
    def __init__(self, outputdir, timesteps=None, embands=None,
                snowbands=None, mask=None, time_start=None, time_end=None,
                wy=None):
        """
        Inputs:
        outputdir - abosulte path to location of outputs
        timestep - list of time steps to return, default all
        embands - list of bands numbers to grab, default all
        snowbands - list of bands numbers to grab, default all
        time_start - water year hour to start grabbing files
        time_end - water year hour to end grabbing files
        wy - water year for finding dates from ipw file wyhrs
        """

        # parse innputs
        self.outputdir = outputdir
        self.timesteps = timesteps
        self.mask = mask
        self.time_start = time_start
        self.time_end = time_end
        self.wy = wy

        # list of band numbers possible
        emnums = range(10)
        snownums = range(9)
        self.embands = embands
        self.snowbands = snowbands
        # if no bands specified, get them all
        if self.embands is None:
            self.embands = emnums
        if self.snowbands is None:
            self.snowbands = snownums

        emnames = ['net_rad', 'sensible_heat', 'latent_heat', 'snow_soil',
                    'precip_advected', 'sum_EB', 'evaporation',
                    'snowmelt', 'SWI', 'cold_content'
                  ]
        snownames = ['thickness', 'snow_density', 'specific_mass', 'liquid_water',
                    'temp_surf', 'temp_lower', 'temp_snowcover',
                    'thickness_lower', 'water_saturation'
                    ]

        self.emdict = {}
        self.snowdict = {}
        for idn, nm in enumerate(emnames):
            self.emdict[emnums[idn]] = nm
        for idn, nm in enumerate(snownames):
            self.snowdict[snownums[idn]] = nm

        # read in and store the data
        self.snow_data, self.em_data = self.read_isnobal_outputs()

    def read_isnobal_outputs(self):
        """
        Inputs:
        self

        Returns:
        snow_data - dictionary of time series spatial data np arrays as well as a
                    time series array
        em_date - dictionary of time series spatial data np arrays as well as a
                    time series array
        """

        # check for errors in time inputs
        if self.timesteps is not None and (self.time_start is not None or self.time_end is not None):
            print('timesteps: {}, time_start: {}, time_end: {}'.format(self.timesteps, self.time_start, self.time_end))
            raise ValueError('Incorrect combination of time inputs for reading files')
        if (self.time_start is None and self.time_end is not None) or (self.time_start is None and self.time_end is not None):
            print('time_start: {}, time_end: {}'.format(self.time_start, self.time_end))
            raise ValueError('Incorrect combination of time inputs for reading files')

        snow_data, em_data = self.read_isnobal_netcdf()

        return snow_data, em_data

    def read_isnobal_netcdf(self):
        """
        Read in specific timesteps and bands from ipw files output from isnobal.

        Input:
        self

        Returns:
        snow_data - dictionary of time series spatial data np arrays as well as a
                    time series array
        em_date - dictionary of time series spatial data np arrays as well as a
                    time series array
        """

        warnings.simplefilter("ignore")

        name_snow_fp = 'snow'
        name_em_fp = 'em'
        # path to snow and em files
        pathsnow = os.path.join(self.outputdir, name_snow_fp+'.nc')
        pathem = os.path.join(self.outputdir, name_em_fp+'.nc')
        # read datasets
        ds_snow = Dataset(pathsnow, 'r')
        ds_em = Dataset(pathem, 'r')

        # hack for different swi names
        if 'runoff' in ds_em.variables:
            self.emdict[8] = 'runoff'

        # get time array from dataset
        ny = len(ds_snow.variables['y'][:])
        nx = len(ds_snow.variables['x'][:])

        # find dates the right way
        nc_time = ds_snow.variables['time'][:]
        t_units = ds_snow.variables['time'].units
        nc_calendar = ds_snow.variables['time'].calendar
        nc_dates = nc.num2date(nc_time, t_units, nc_calendar)
        self.dates = nc_dates

        # calculate water year hours
        time = np.zeros(len(nc_dates))
        for ix, t in enumerate(nc_dates):
            # calculate new date
            dtt = t
            # convert to wyhr
            wyhr = calculate_wyhr_from_date(dtt)
            # store wyhr in time array
            time[ix] = wyhr

        # mask is ones if none specified
        if self.mask is None:
            mask = np.ones((ny,nx))
        else:
            mask = self.mask

        # empty dictionaries to store numpy arrays of each variable
        snow_data = {}
        em_data = {}

        # print('Processing datasets')
        # parse user input timesteps or use all
        if self.timesteps is None and self.time_start is None and self.time_end is None:
            timesteps = time
            itd = len(time)
            for ebd in self.embands:
                e_name = self.emdict[ebd]
                em_data[ebd] = ds_em.variables[e_name][:]*mask
                #em_data[ebd][em_data[ebd] <= -75.0] = np.nan
            for sbd in self.snowbands:
                s_name = self.snowdict[sbd]
                snow_data[sbd] = ds_snow.variables[s_name][:]*mask
                #snow_data[sbd][snow_data[sbd] <= -75.0] = np.nan

        # if given a start and stop time
        elif self.timesteps is None and self.time_start is not None and self.time_end is not None:
            # only time steps in desired timeframe
            timesteps = time[(time >= self.time_start) & (time <= self.time_end)]
            itd = timesteps

            # pull data out at specified times
            for ebd in self.embands:
                # find variable name
                e_name = self.emdict[ebd]

                em_data[ebd] = np.zeros((len(itd), ny, nx))
                for ide, it in enumerate(itd):
                    time_samp = np.where(time == it)[0]
                    tmp_em_data = ds_em.variables[e_name][time_samp,:]
                    #tmp_em_data[np.isnan(tmp_em_data)] = 0.0
                    em_data[ebd][ide,:] = tmp_em_data*mask

            for sbd in self.snowbands:
                # find variable name
                s_name = self.snowdict[sbd]
                snow_data[sbd] = np.zeros((len(itd), ny, nx))
                for ids, it in enumerate(itd):
                    time_samp = np.where(time == it)[0]
                    tmp_snow_data = ds_snow.variables[s_name][time_samp,:]
                    if sbd in [4, 5, 6]:
                        tmp_snow_data[tmp_snow_data == -75.0] = np.nan
                    #tmp_snow_data[np.isnan(tmp_snow_data)] = 0.0

                    snow_data[sbd][ids,:] = tmp_snow_data*mask

        # if given a list of timesteps
        elif self.timesteps is not None:
            timesteps = np.array(self.timesteps)
            # find indices that have matching times
            itd = []
            for it, t in enumerate(time):
                if t in timesteps:
                    itd.append(it)
            if len(itd) == 0:
                raise ValueError('No matching timesteps to those given')

            itd = np.array(itd)

            # pull data out at specified times
            for ebd in self.embands:
                # find variable name
                e_name = self.emdict[ebd]

                em_data[ebd] = np.zeros((len(itd), ny, nx))
                for ide, it in enumerate(itd):
                    em_data[ebd][ide,:] = ds_em.variables[e_name][it,:]*mask
                    #em_data[ebd][ide,:][em_data[ebd][ide,:] <= -75.0] = np.nan

            for sbd in self.snowbands:
                # find variable name
                s_name = self.snowdict[sbd]
                snow_data[sbd] = np.zeros((len(itd), ny, nx))
                for ids, it in enumerate(itd):
                    snow_data[sbd][ids,:] = ds_snow.variables[s_name][it,:]*mask
                    #snow_data[sbd][ids,:][snow_data[sbd][ids,:] <= -75.0] = np.nan

        # store timesteps
        self.time = timesteps
        # close datasets
        ds_snow.close()
        ds_em.close()

        return snow_data, em_data




import os
import pandas as pd
from sqlalchemy import and_
from snowav.database.database import connect, make_session
from snowav.database.tables import RunMetadata, Results, VariableUnits
from sys import exit

def query(self):
    '''
    Query snow server snowav database. Requires correct database login
    information in the config file [database] section.

    See README.md and CoreConfig.ini for details and more information.

    '''

    print('\nConfig option [query] = True, running database query...')

    if ((self.q_basins is None) or (self.q_run_name is None) or
        (self.q_start_date is None) or (self.q_end_date is None)):
        raise Exception('Required config [query] fields missing, '
                        'see CoreConfig.ini')

    if type(self.q_basins) != list:
        basins = [self.q_basins]
    else:
        basins = self.q_basins

    bdict, cnx, out = connect(sqlite = self.sqlite, sql = self.mysql,
                              plotorder = basins, user = self.db_user,
                              password = self.db_password, host = self.db_host,
                              port = self.db_port, convert = self.db_convert)

    value = self.q_value
    run_name = self.q_run_name
    start_date = self.q_start_date
    end_date = self.q_end_date
    total = self.q_total
    csv_base_path = self.q_csv_base_path
    output = self.q_output
    db = self.q_database
    print_runs = self.q_print_all_runs

    # Make df from database
    db_val = pd.DataFrame(columns = basins)
    session = make_session(self.q_database)

    for bid in basins:

        # Get available run_names for the watershed
        wid = int(bdict[basins[0]]['watershed_id'])

        # print out all available runs for the time period if desired
        if print_runs is True:

            q = session.query(RunMetadata).join(Results).filter(and_(RunMetadata.watershed_id == wid),
                                                                    (Results.date_time >= start_date),
                                                                    (Results.date_time <= end_date))
            df = pd.read_sql(q.statement, q.session.connection())
            names = df.run_name.unique()

            # run_name and run_id
            rid = {}
            for name in names:
                rid[name] = df[df['run_name'] == name]['run_id'].values

            qry = session.query(Results).join(RunMetadata).filter(and_((Results.date_time >= start_date),
                                                      (Results.date_time <= end_date),
                                                      (RunMetadata.run_name.in_(names)),
                                                      (RunMetadata.watershed_id == wid),
                                                      (Results.basin_id == int(bdict[bid]['basin_id']))))

            df = pd.read_sql(qry.statement, qry.session.connection())

            qry2 = session.query(RunMetadata).join(Results).filter(and_((Results.date_time >= start_date),
                                                      (Results.date_time <= end_date),
                                                      (RunMetadata.run_name.in_(names)) ))
            df2 = pd.read_sql(qry2.statement, qry2.session.connection())

            dash = '-'*100
            print('\n{}'.format(dash))
            print('On snowav database for {}:'.format(bid))
            print(dash)

            for name in rid:
                try:
                    stime = df[df['run_id'].isin(rid[name])]['date_time'].min().date().strftime("%Y-%-m-%-d")
                except:
                    stime = 'nan'

                try:
                    etime = df[df['run_id'].isin(rid[name])]['date_time'].max().date().strftime("%Y-%-m-%-d")
                except:
                    etime  = 'nan'

                print('{:<25s}{:<25s}{:<25s}'.format(name, stime, etime ))

            print('{}\n'.format(dash))

        bid_name = bid.replace(" ", "_")
        csv_file = '{}{}_{}_{}.csv'.format(csv_base_path,bid_name,run_name,value)
        qry = session.query(Results).join(RunMetadata).filter(and_(
                        (Results.date_time >= start_date),
                        (Results.date_time <= end_date),
                        (RunMetadata.run_name == run_name),
                        (RunMetadata.watershed_id == wid),
                        (Results.variable == value),
                        (Results.basin_id == int(bdict[bid]['basin_id']))))

        df = pd.read_sql(qry.statement, qry.session.connection())

        # Use the first call to get the units
        qry2 = session.query(VariableUnits).filter(VariableUnits.id == int(df['variable_id'].values[0]))
        df2 = pd.read_sql(qry2.statement, qry2.session.connection())
        df['units'] = df2['unit'].values[0]

        if total is True:
            fqry = df[df['elevation'] == 'total']

        else:
            fqry = df

        fqry = fqry.sort_values(by=['date_time'])

        db_val.loc['total',bid] = fqry['value'].values

        cols = ['date_time','value','units','elevation']

        if output == 'print':
            with pd.option_context('display.max_rows', None, 'display.max_columns', None):
                print('\n\nsnowav database values \n'
                      'basin: {}\n'
                      'run_name: {}\n'
                      'value: {}\n\n'
                      '{}'.format(bid, run_name, value, fqry[cols]))

        if output == 'csv':
            if total:
                fqry.to_csv(csv_file)
            else:
                for e in fqry.elevation.unique().tolist():
                    banded = fqry.copy()
                    banded = banded.set_index('date_time')
                    banded = banded[banded['elevation'] == e]
                    banded = pd.DataFrame(banded['value'].rename('{} {} [ft]'.format(value, e)))
                    banded.to_csv('{}_{}.csv'.format(csv_file.split('.csv')[0],e))

    session.close()
    print('\nQuery complete, exiting snowav. To process results, set [query] query: False\n')
    exit()



from datetime import datetime
import pandas as pd

from snowav.framework.query import query
from snowav.config.config import UserConfig
from snowav.framework.process import Process
from snowav.framework.figures import figures
from snowav.report.report import report
from snowav.database.database import run_metadata
from snowav.inflow.inflow import excel_to_csv


class Snowav(object):

    def __init__(self, config_file=None, awsm=None, end_date=None):
        """ Read config file, parse config options, process results, put
        results on database, make figures, and make pdf report.

        Args
        -----
        config_file {str}: config file
        awsm {class}: awsm class
        end_date {str}: overwrite of config end_date
        """

        if end_date is not None:
            try:
                end_date = pd.to_datetime(end_date)
            except Exception as e:
                print(e)

        # get and parse config options
        cfg = UserConfig(config_file, awsm=awsm, end_date=end_date)
        cfg.parse()

        # query existing database without processing
        if cfg.query_flag:
            query(cfg)

        # put run metadata on database
        run_metadata(cfg)

        # process
        process = Process(cfg)

        if cfg.inflow_flag and cfg.inflow_data is not None:
            args = {'path': cfg.inflow_data,
                    'csv_file': cfg.summary_csv,
                    'basin_headings': cfg.basin_headings,
                    'inflow_headings': cfg.inflow_headings,
                    'file_base': cfg.file_base,
                    'sheet_name': cfg.sheet_name,
                    'skiprows': cfg.skiprows,
                    'date_idx': cfg.date_idx,
                    'wy': cfg.wy,
                    'overwrite': cfg.overwrite,
                    'convert': cfg.convert}

            excel_to_csv(args, cfg._logger)

        figures(cfg, process)

        # Do additional processing and figures if forecast is supplied. Some
        # field will be overwritten during forecast processing
        if cfg.forecast_flag:
            cfg._logger.info(' Starting forecast processing...')

            run_metadata(cfg, cfg.for_run_name)

            cfg.pargs['run_id'] = cfg.run_id
            cfg.pargs['vid'] = cfg.vid
            flags, out, pre, rain, density = process(cfg.pargs)

            for log in out:
                cfg._logger.info(log)

            cfg.density = density
            cfg.rain_total = rain
            cfg.precip_total = pre

            # figures for forecast run
            figures(cfg)

        if cfg.report_flag:
            report(cfg, process)

        elapsed = str(datetime.now() - cfg.proc_time_start)

        cfg._logger.info(' Completed snowav processing, '
                         'elapsed time: {}'.format(elapsed))



import os
import datetime
from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm import backref
from snowav.database.tables import Base, RunMetadata, Watershed, Basin, Results, VariableUnits, Watersheds, Basins
from sqlalchemy import and_
import pandas as pd
try:
    import smrf
except:
    print('Could not import smrf')
try:
    import awsm
except:
    print('Could not import awsm')
import snowav
from snowav import database
import numpy as np
import copy
import urllib.parse
import mysql.connector


'''
mysql -h 172.17.0.2 -P 3306 -u mark -p

snowav -b 'basin' -wy wy -db mysql+mysqlconnector://<user>:<pwd>@172.17.0.2/snowav

'''

run_name = 'sj_wy2019_forecast'
start_date = datetime.datetime(2018,10,1)
end_date = datetime.datetime(2019,9,30)
# basins = ['Middle Fork','West Kings','Dinkey Creek','Middle South Fork','South Fork','Mill Creek','North Fork']
basins = ['San Joaquin River Basin','South Fork','Main','Auberry','Redinger']

user = 'user'
pwd = 'password'
host = '172.17.0.2'
port = '3306'
db = 'snowav'

db_engine = 'mysql+mysqlconnector://{}:{}@{}:{}/{}'.format(user,pwd,host,port,db)
engine = create_engine(db_engine)
Base.metadata.create_all(engine)
DBSession = sessionmaker(bind=engine)
session = DBSession()

for bid in basins:
    wid =  Basins.basins[bid]['watershed_id']

    qry = session.query(Results).join(RunMetadata).filter(and_(
                    (Results.date_time >= start_date),
                    (Results.date_time <= end_date),
                    (RunMetadata.run_name == run_name),
                    (Results.basin_id == Basins.basins[bid]['basin_id'])))

    df = pd.read_sql(qry.statement, qry.session.connection())

    # Delete Results by date range, run_id, and basin
    for r in df.run_id.unique():
        session.query(Results).filter(and_(
            (Results.date_time >= start_date),
            (Results.date_time <= end_date),
            (Results.run_id == int(r)),
            (Results.basin_id == Basins.basins[bid]['basin_id']))).delete()

        session.flush()

    # Query Metadata to see what's remaining
    qry = session.query(RunMetadata).join(Results).filter(and_(RunMetadata.watershed_id == wid),
                                                        (RunMetadata.run_name == run_name),
                                                        (Results.date_time >= start_date),
                                                        (Results.date_time <= end_date))


    df = pd.read_sql(qry.statement, qry.session.connection())

    # run_name and run_id
    rid = {}
    rid[run_name] = df[df['run_name'] == run_name]['run_id'].values

    # Toggling

    # for n in rid[run_name]:
    #     session.query(RunMetadata).filter(RunMetadata.run_id == int(n)).delete()
    #     session.flush()
    #     session.commit()


    # for n in rid[run_name]:
    #         session.query(VariableUnits).filter(VariableUnits.run_id == int(n)).delete()
    #         session.flush()
    #         session.commit()

session.close()



import coloredlogs
from datetime import datetime
import logging
import netCDF4 as nc
import numpy as np
import os
import pandas as pd
from sys import exit
import utm

from inicheck.config import MasterConfig
from inicheck.output import print_config_report
from inicheck.tools import get_user_config, check_config, cast_all_variables
import snowav
from snowav.database.models import AwsmInputsOutputs
from snowav.database.database import Database


class PointValues(object):
    """ Point values processing to pull time-series model results from
    specific pixels.

    If config files aren't passed, the class contains only hard-coded string
    definitions.

    NOTE: Class attributes are automatically assigned from the master config
    file.

    Args
    ------
    master_config: {string} path to inicheck master config
    config_file: {sring} path to config file
    """

    def __init__(self, master_config=None, config_file=None):

        # run processing if both are passed, otherwise PointValues
        # contains only the hard-coded string definitions
        if sum([x is not None for x in [master_config, config_file]]) == 2:

            # inicheck the config file
            mcfg = MasterConfig(path=master_config)
            ucfg = get_user_config(config_file, mcfg=mcfg)
            ucfg.apply_recipes()
            ucfg = cast_all_variables(ucfg, ucfg.mcfg)

            # check config, warn and exit if necessary
            warnings, errors = check_config(ucfg)
            if len(errors) > 0 or len(warnings) > 0:
                print_config_report(warnings, errors)
                if len(errors) > 0:
                    exit()

            # get a list of master config fields to check assignment of fields
            # in PointValues class
            mcfg_fields = []
            for section, d in ucfg.mcfg.cfg.items():
                mcfg_fields += list(d.keys())

            cfg = ucfg.cfg

            # Add in class attributes from config file
            for section, d in cfg.items():
                for item, v in d.items():

                    # assign if it exists in the master config
                    if item in mcfg_fields:
                        setattr(self, item, v)

            self.create_log()
            self.logger.info(' Starting point values processing...')

            # additional checks for user config options below...

            # check sqlite base path and file extension
            if self.database is not None:
                if not os.path.exists(os.path.dirname(self.database)):
                    self.logger.error(" [point_values] database: '{}' base "
                                      "path not valid".format(self.database))
                    raise OSError("invalid path -> {}".format(self.database))

                if os.path.splitext(self.database)[1] != '.db':
                    self.logger.error(" [point_values] database: '{}' "
                                      "must be '.db' "
                                      "file".format(self.database))
                    raise OSError("file type must be '.db'")
                self.db_type = 'sqlite'

            # if sql database, make sure all credentials exist
            if self.database is None:
                input_list = [self.user, self.password, self.host, self.port]
                if sum([x is not None for x in input_list]) != 4:
                    self.logger.warning(" User has not supplied all database "
                                        "credentials in [point_values]")
                self.db_type = 'sql'

            # csv saving requires directory
            if self.csv_output and self.csv_output_dir is None:
                self.logger.error(" [point_values] csv_output: True, must "
                                  "supply csv_output_dir field")
                exit()

            self.start_date_str = self.start_date.date().strftime("%Y%m%d")
            self.end_date_str = self.end_date.date().strftime("%Y%m%d")
            self.begin_time = datetime.now()

            snowav_properties = AwsmInputsOutputs()
            all_properties = (snowav_properties.smrf_variables +
                              snowav_properties.awsm_variables)

            # check and assign properties
            self.properties_lookup = {}

            if self.properties == ['all']:
                self.properties = all_properties
                if 'precip_z' in self.properties:
                    self.properties.remove('precip_z')
            else:
                for p in self.properties:
                    if p not in all_properties:
                        self.logger.error(' Provided [basin] properties: '
                                          '"{}" not a valid property, must be '
                                          'in {}'.format(p, all_properties))

            self.bandsmap = {}
            for p in all_properties:
                if p != 'precip_z':
                    self.bandsmap[p] = snowav_properties.vars[p]['nc_name']

            # make a lookup dict that is {file: [variables]} so we know what to
            # load for each file
            files = []
            var_list = []
            for p in snowav_properties.vars.keys():
                file = snowav_properties.vars[p]['nc_name']
                if (file is not None and
                        snowav_properties.vars[p]['file'] not in files):
                    files.append(snowav_properties.vars[p]['file'])

            for file in files:
                if file not in self.properties_lookup.keys():
                    self.properties_lookup[file] = {}
                    self.properties_lookup[file]['variables'] = []
                    self.properties_lookup[file]['bands'] = {}

            self.unit_map = {}
            for f in self.properties_lookup.keys():
                for v in self.properties:
                    variable = self.bandsmap[v]

                    self.unit_map[variable] = \
                        snowav_properties.vars[v]['units']

                    if f == snowav_properties.vars[v]['file']:
                        self.properties_lookup[f]['variables'].append(v)
                        band = snowav_properties.vars[v]['band']
                        self.properties_lookup[f]['bands'][v] = band

                        # make list for the data dfs - these are the headings

                        # The variable names snow.nc and snow_density.nc are
                        # both 'snow_density', this is an issue!
                        # This is also addressed in load_data() and save_csv()
                        if v == 'density':
                            var_list.append('density')
                        else:
                            var_list.append(variable)

            self.logger.debug(" Variables: {}".format(var_list))

            # these will be converted to dicts with location keys later
            self.tvar = pd.DataFrame(columns=var_list)

        # hard coded strings relating to AWSM and PointValues csv file
        # NOTE that database Pixels.location field is created by:
        #    os.path.abspath(os.path.join(pv.run_dirs[0], '..', '..'))
        # in load_data()
        self.smrf_dir = 'smrfOutputs'
        self.data_str = '/data'
        self.datadata_str = '/data/data'
        self.runs_str = '/runs'
        self.runsrun_str = '/runs/run'
        self.name_col = 'name'
        self.description_col = 'description'
        self.model_xind = 'model xind'
        self.model_yind = 'model yind'
        self.index_columns = [self.model_xind, self.model_yind]
        self.location_columns = ['latitude', 'longitude']
        self.csv_cols = [self.name_col,
                         self.description_col,
                         self.model_xind,
                         self.model_yind] + self.location_columns

    def assign_vars(self, kwargs):
        """ Assign PointValues attributes.

        Args
        ------
        kwargs {dict}: {item, value}
        """

        if not isinstance(kwargs, dict):
            raise TypeError("kwargs must be dict")

        for item, v in kwargs.items():
            setattr(self, item, v)

    def create_log(self):
        """ Create logger. """

        level_styles = {'info': {'color': 'white'},
                        'error': {'color': 'red'},
                        'debug': {'color': 'green'},
                        'warning': {'color': 'yellow'}}

        field_styles = {'hostname': {'color': 'magenta'},
                        'programname': {'color': 'cyan'},
                        'name': {'color': 'white'},
                        'levelname': {'color': 'white', 'bold': True},
                        'asctime': {'color': 'green'}}

        # start logging
        log_level = self.log_level.upper()

        numeric_level = getattr(logging, log_level, None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % log_level)

        fmt = '%(levelname)s:%(name)s:%(message)s'
        logging.basicConfig(level=numeric_level)
        coloredlogs.install(level=numeric_level,
                            fmt=fmt,
                            level_styles=level_styles,
                            field_styles=field_styles)

        self.assign_vars({'logger': logging.getLogger(__name__)})


def get_paths(pv):
    """ Get the 'data' and 'runs' directories for inputs and outputs.

        ../runs/runYYYYMMDD
        ../data/dataYYYYMMDD/smrfOutputs

    Assigns:
    pv.data_dirs
    pv.runs_dirs

    Args
    ------
    pv {class} PointValues class
    """

    # get all run dirs
    run_path = pv.directory
    r = os.listdir(run_path)
    run_dirs = [os.path.join(run_path, x) for x in r]
    run_dirs.sort()

    for dd in run_dirs:
        pv.logger.debug(" Checking run dir: {}".format(dd))

    # get all data dirs
    data_path = run_path.replace(pv.runs_str, pv.data_str)
    d = os.listdir(data_path)
    d_dirs = [os.path.join(x, pv.smrf_dir) for x in d]
    data_dirs = [os.path.join(data_path, x) for x in d_dirs]
    data_dirs.sort()

    for dd in run_dirs:
        pv.logger.debug(" Checking data dir: {}".format(dd))

    if len(run_dirs) != len(data_dirs):
        pv.logger.error(' There are mismatched number of subdirectories '
                        'in {} and {}'.format(run_path, data_path))

    # filter do the ones in requested date range
    if sum([x is not None for x in [pv.start_date, pv.end_date]]) == 2:
        pv.logger.debug(' Parsing folder dates...')
        pv.data_dirs = []
        pv.run_dirs = []

        for d in data_dirs:
            s = d.split(pv.datadata_str)[1].split('/' + pv.smrf_dir)[0]
            dt = datetime(int(s[:4]), int(s[4:6]), int(s[-2::]))

            if pv.start_date <= dt <= pv.end_date:
                pv.data_dirs.append(d)
                pv.logger.info(' Using data dir: {}'.format(d))

        for r in run_dirs:
            s = r.split(pv.runsrun_str)[1]
            dt = datetime(int(s[:4]), int(s[4:6]), int(s[-2::]))
            if pv.start_date <= dt <= pv.end_date:
                pv.run_dirs.append(r)
                pv.logger.info(' Using run dir: {}'.format(r))

        # check again in case the parsing changed anything
        if len(run_dirs) != len(data_dirs):
            pv.logger.error(' There are mismatched number of subdirectories '
                            'in {} and {}'.format(run_path, data_path))


def get_basin_info(pv):
    """ Load topo.nc file and get basin info.

    Args
    ------
    pv {class} PointValues class
    """

    pv.logger.debug(" Loading basin info from {}".format(pv.topo))

    # load topo.nc file
    try:
        data = nc.Dataset(pv.topo)
    except Exception as e:
        print(e)

    basin_name = data.variables['mask'].long_name
    pv.basin_name = ''.join(basin_name.split())
    pv.basin_x = len(data.variables['x'])
    pv.basin_y = len(data.variables['y'])
    pv.utm_x = data.variables['x'][:]
    pv.utm_y = data.variables['y'][:]
    pv.dem = data.variables['dem'][:]
    data.close()

    pv.logger.debug(" Basin y: {}".format(pv.basin_y))
    pv.logger.debug(" Basin x: {}".format(pv.basin_x))


def get_locations(pv):
    """ Get location data from csv file. If using latitude and longitude,
    convert to model x and y indices.

    Assigns:
    pv.data

    Args
    ------
    pv {class} PointValues class
    """

    pv.logger.debug(" Loading data for locations from "
                    "{}".format(pv.locations_csv))

    # load csv
    try:
        data = pd.read_csv(pv.locations_csv)
    except Exception as e:
        print(e)

    if data.empty:
        pv.logger.error(" {} is empty ".format(pv.locations_csv))

    columns = list(data.columns)

    if pv.name_col not in columns:
        pv.logger.warning(" '{}' column is missing".format(pv.name_col))

    if pv.description_col not in columns:
        pv.logger.warning(" '{}' column is missing".format(pv.description_col))

    # check if df contains complete set of either headings
    check = True

    # index
    if all(e in columns for e in pv.index_columns):
        pv.logger.debug(" Checking file {} "
                        "headings".format(pv.index_columns))

        if np.isnan(data[pv.index_columns].values).any() > 0:
            pv.logger.debug(" {} may contain empty "
                            "fields".format(pv.index_columns))
        else:
            data = data[pv.index_columns + [pv.name_col] + [pv.description_col]]
            check = False
            pv.logger.debug(" Using data in "
                            "{}".format(pv.index_columns))

            # check that indices are within basin dimensions
            if not all(x >= 0 for x in list(data[pv.model_xind].values)):
                pv.logger.error(" csv file contains index value <0")
                raise ValueError("Index out of bounds")

            if not all(x <= pv.basin_x for x in
                       list(data[pv.model_xind].values)):
                pv.logger.error(" csv file contains index value "
                                ">{}".format(pv.basin_x))
                raise ValueError("Index out of bounds")

            if not all(x >= 0 for x in list(data[pv.model_yind].values)):
                pv.logger.error(" csv file contains index value <0")
                raise ValueError("Index out of bounds")

            if not all(x <= pv.basin_y for x in
                       list(data[pv.model_yind].values)):
                pv.logger.error(" csv file contains index value "
                                ">{}".format(pv.basin_y))
                raise ValueError("Index out of bounds")

    # latitude and longitude
    if all(e in columns for e in pv.location_columns) and check:
        pv.logger.debug(" Checking file {} "
                        "headings".format(pv.location_columns))

        if np.isnan(data[pv.location_columns].values).any() > 0:
            pv.logger.warning(" {} may contain empty "
                              "fields".format(pv.location_columns))
        else:
            check = False
            pv.logger.debug(" Using data in "
                            "{}".format(pv.location_columns))

        # convert to indices and assign
        for i, (index, row) in enumerate(data.iterrows()):
            ll = utm.from_latlon(row['latitude'], row['longitude'])
            xind = np.where(abs(pv.utm_x - ll[0]) ==
                            min(abs(pv.utm_x - ll[0])))[0]
            yind = np.where(abs(pv.utm_y - ll[1]) ==
                            min(abs(pv.utm_y - ll[1])))[0]

            if xind == 0 or xind == pv.basin_x - 1:
                pv.logger.warning(" For {} xind = {}, "
                                  "may be out of basin".format(ll, xind))

            if yind == 0 or yind == pv.basin_x - 1:
                pv.logger.warning(" For {} yind = {}, "
                                  "may be out of basin".format(ll, yind))

            # assign x, y, name, and description
            data.loc[i, pv.model_xind] = int(xind)
            data.loc[i, pv.model_yind] = int(yind)

    if check:
        pv.logger.error(" csv must contain set of either {} or {} "
                        "columns".format(pv.index_columns,
                                         pv.location_columns))

    pv.data = data.copy()


def save_csv(pv):
    """ Save results in csv format.

    Args
    ------
    pv {class} PointValues class
    """

    now_str = datetime.now().date().strftime("%Y-%m-%d")
    date_col = 'Date generated: {}'.format(now_str)

    header = ['USDA Agicultural Research Service AWSM Model Data',
              'Data provided are SMRF inputs (hourly) and iSnobal outputs (daily)',
              date_col,
              'Contact: Scott Havens <scott.havens@usda.gov>\n']

    for loc in pv.var_dict.keys():
        df = pv.var_dict[loc]['data']
        df.sort_index(inplace=True)
        df.index.name = 'date_time'

        # add units to the headings
        columns = list(df.columns)

        for c in columns:
            # prevent renaming to snow_density
            if c == 'density':
                rdict = {c: "density [kg/m^3]"}
            else:
                rdict = {c: "{} [{}]".format(c, pv.unit_map[c])}
            df = df.rename(columns=rdict)

        loc_str = "x{}_y{}".format(str(int(loc[1])), str(int(loc[0])))
        name = "{}_{}_{}_variables_{}.csv".format(pv.basin_name,
                                                  pv.start_date_str,
                                                  pv.end_date_str,
                                                  loc_str)

        # df = df.sort_index(inplace=True)

        filename = os.path.join(pv.csv_output_dir, name)

        with open(filename, mode='w', encoding='utf-8') as f:
            f.write('\n'.join(header))

        df.to_csv(filename, encoding='utf-8', mode='a')

        pv.logger.info(' Saved: {}'.format(filename))


def load_data(pv):
    """ Load smrf and iSnobal outputs from pv.run_dirs and pv.data_dirs.

    Assigns:
    pv.var_dict {dict}: {(x, y): 'data': df,
                                 'name': name,
                                 'description': description,
                                 'location': location}

    where df.columns = ['air_temp','vapor_pressure',...]

    Args
    ------
    pv {class}: PointValues class
    """

    base_location = os.path.abspath(os.path.join(pv.run_dirs[0], '..', '..'))

    pv.var_dict = {}
    for index, row in pv.data.iterrows():
        loc = (int(row[pv.model_xind]), int(row[pv.model_yind]))
        pv.var_dict[loc] = {'data': pv.tvar.copy(),
                            'name': row[pv.name_col],
                            'description': row[pv.description_col],
                            'location': base_location}

    # each runs/ and data/ directory
    for n, d in enumerate(pv.run_dirs + pv.data_dirs):
        pv.logger.debug(" Working in {}".format(d))

        # each file, i.e. snow.nc
        for file in os.listdir(d):
            filepath = os.path.join(d, file)

            # if it's a valid file and one we need
            if os.path.isfile(filepath) and \
                    file in list(pv.properties_lookup.keys()):

                try:
                    data = nc.Dataset(filepath)
                    date_time = nc.num2date(data.variables['time'][:],
                                            data.variables['time'].units)
                except Exception as e:
                    print(e)

                pv.logger.debug(" Opened {}".format(file))

                # ensure that outputs are hourly
                for i, dt in enumerate(date_time):
                    date_time[i] = date_time[i].replace(minute=0,
                                                        second=0,
                                                        microsecond=0)

                # for each band
                for v, b in pv.properties_lookup[file]['bands'].items():

                    # for each location
                    for item, row in pv.data.iterrows():
                        loc = (int(row[pv.model_xind]), int(row[pv.model_yind]))

                        # snowav --> .nc
                        if v in pv.bandsmap.keys() and b is not None:
                            v = pv.bandsmap[v]

                        try:
                            # nans in some of the netcdf files give runtime
                            # warnings when being loaded. As far as I can tell
                            # the only fix is fill_value when creating.
                            img = data.variables[v][:]
                        except KeyError as e:
                            print('{} not a variable in {}'.format(e, file))

                        # check dimensions of the first image opened
                        if n == 0 and (pv.basin_y, pv.basin_x) != img[0].shape:
                            pv.logger.error(" output dimensions {} "
                                            "do not match topo "
                                            "({}, {})".format(img[0].shape,
                                                              pv.basin_y,
                                                              pv.basin_x))

                        for ih in range(0, len(date_time)):
                            if v == 'snow_density' and file == 'snow.nc':
                                vt = 'density'
                            else:
                                vt = v

                            if ih < 24:
                                dth = date_time[ih]
                                pv.var_dict[loc]['data'].loc[dth, vt] = \
                                    img[ih, loc[1], loc[0]]

                                pv.logger.debug(" Write to df: {}, {}, "
                                                "{}".format(v, loc, dth))

                data.close()

    pv.var_dict[loc]['data'].sort_index(inplace=True)


def put_on_database(db, pv):
    """ Database stream for checking for existing results, metdata, and
    results.

    Args
    ------
    db {class}: Database class
    pv {class}: PointValues class
    """

    # for each location
    for pt in pv.var_dict:

        # collect metadata
        metadata = {
            'location': str(pv.var_dict[pt]['location']),
            'model_row': int(pt[0]),
            'model_col': int(pt[1]),
            'description': str(pv.var_dict[pt]['description']),
            'name': str(pv.var_dict[pt]['name']),
            'utm_x': float(pv.utm_x[pt[0]]),
            'utm_y': float(pv.utm_y[pt[1]]),
            'elevation': float(pv.dem[pt[1], pt[0]])
        }

        params = {
            'Pixels': ('model_row', '==', int(pt[0])),
            'Pixels': ('model_col', '==', int(pt[1])),
            'Pixels': ('name', '==', str(pv.var_dict[pt]['name'])),
            'Pixels': ('location', '==', str(pv.var_dict[pt]['location'])),
            'Pixels': ('description', '==', str(pv.var_dict[pt]['description'])),

            # not being filtered inside query() at the moment
            'PixelsData': ('date_time', '==', datetime(2020, 3, 3, 23, 0))
        }

        pixelsparams = {
            'Pixels': ('model_row', '==', int(pt[0])),
            'Pixels': ('model_col', '==', int(pt[1])),
            'Pixels': ('name', '==', str(pv.var_dict[pt]['name'])),
            'Pixels': ('location', '==', str(pv.var_dict[pt]['location'])),
            'Pixels': ('description', '==', str(pv.var_dict[pt]['description'])),
        }

        # first, see if records already exist with the same values
        results = db.query(params, logger=pv.logger)

        # filter outside of query for now
        results = results[(results.model_row == pt[0]) &
                          (results.model_col == pt[1]) &
                          (results.name == str(pv.var_dict[pt]['name'])) &
                          (results.date_time >= pv.start_date) &
                          (results.date_time <= pv.end_date)]
        existing_record_id = pd.unique(results['id'])

        # if no existing records, put on database
        if results.empty:

            # insert metadata
            db.insert('Pixels', metadata, logger=pv.logger)
            pv.logger.debug(" Metadata to database for point: "
                            "({},{})".format(metadata['model_row'],
                                             metadata['model_col']))

            # get associated Pixels.id metadata for PixelsData
            results = db.query(pixelsparams, logger=pv.logger)

            # for this case we can't filter by 'date_time', because there
            # aren't any records yet
            results = results[(results.model_row == pt[0]) &
                              (results.model_col == pt[1]) &
                              (results.name == str(pv.var_dict[pt]['name']))]

            max_record_id = int(max(pd.unique(results['id'])))

            # prepare data to put on database
            df = pv.var_dict[pt]['data']

            for idx, row in df.iterrows():
                row = row.astype('float')
                p = {'pixel_id': max_record_id,
                     'date_time': idx}
                res = row.to_dict()
                res = {k: res[k] for k in res if not np.isnan(res[k])}
                pixeldata = {**p, **res}

                # insert data
                db.insert('PixelsData', pixeldata, logger=pv.logger)

            # log just one
            pv.logger.debug(" Data to database for point: "
                            "({},{})".format(metadata['model_row'],
                                             metadata['model_col']))

        # if there are existing record, check overwrite
        else:
            if pv.overwrite:
                # first, delete the existing metadata and records based on the
                # existing Pixels.id
                for rec in existing_record_id:
                    # deletedata = {'id': int(rec)}

                    # first, delete the data
                    db.delete('PixelsData',
                              {'pixel_id': int(rec)},
                              logger=pv.logger)

                    results = db.query(params, logger=pv.logger)
                    results = results[(results.model_row == pt[0]) &
                                      (results.model_col == pt[1]) &
                                      (results.name == str(pv.var_dict[pt]['name'])) &
                                      (results.date_time >= pv.start_date) &
                                      (results.date_time <= pv.end_date)]

                    # if that removed everything, also delete the metadata
                    if results.empty:
                        db.delete('Pixels',
                                  {'id': int(rec)},
                                  logger=pv.logger)

                # insert metadata
                db.insert('Pixels', metadata, logger=pv.logger)
                pv.logger.debug(" Metadata to database for point: "
                                "({},{})".format(metadata['model_row'],
                                                 metadata['model_col']))

                # get associated Pixels.id metadata for PixelsData with the
                # same query params
                results = db.query(pixelsparams, logger=pv.logger)

                # filter outside of query for now
                results = results[(results.model_row == pt[0]) &
                                  (results.model_col == pt[1]) &
                                  (results.name == str(pv.var_dict[pt]['name']))]
                record_id = int(max(pd.unique(results['id'])))

                # prepare data to put on database
                df = pv.var_dict[pt]['data']

                for idx, row in df.iterrows():
                    row = row.astype('float')
                    p = {'pixel_id': int(record_id),
                         'date_time': idx}
                    res = row.to_dict()
                    res = {k: res[k] for k in res if not np.isnan(res[k])}
                    pixeldata = {**p, **res}

                    # insert data
                    db.insert('PixelsData', pixeldata, logger=pv.logger)

                # log just one
                pv.logger.debug(" Data to database for point: "
                                "({},{})".format(metadata['model_row'],
                                                 metadata['model_col']))

            else:
                pv.logger.info(" Database records exist for {}, ({},{}) "
                               "and [point_values] overwrite: False, "
                               "values not put on "
                               "database".format(metadata['location'],
                                                 metadata['model_row'],
                                                 metadata['model_col']))


def run_point_values(point_values_config, master_config, blank):
    """ Run point values processing.

    This does a bulk records delete for matching 'location', 'model_row',
    and 'model_col' on the database if existing records are present.

    Args
    ------
    point_values_config {string}: config file
    master_config {string}: point values master config
    blank {bool}: make a blank locations_csv file
    """

    # use default CoreConfig.ini if one is not supplied
    if master_config is None:
        master_config = os.path.abspath(
            os.path.join(snowav.__path__[0],
                         "./config/PointValuesCoreConfig.ini")
        )

    if point_values_config is not None:
        if not os.path.isfile(os.path.abspath(point_values_config)):
            raise Exception('{} not a valid file'.format(master_config))

        if not os.path.isfile(os.path.abspath(master_config)):
            raise Exception('{} not a valid file'.format(master_config))

        # initialize class
        pv = PointValues(master_config, point_values_config)

        # initialize Database class
        db = Database(user=pv.user, password=pv.password, host=pv.host,
                      port=pv.port, db_type=pv.db_type, database=pv.database)

        # make and check database connection
        db.make_connection(logger=pv.logger)

        # check database tables
        db.check_tables(logger=pv.logger)

        # load topo.nc and get basin info
        get_basin_info(pv)

        # get and parse locations from the csv file
        get_locations(pv)

        # get data paths
        get_paths(pv)

        # load data
        load_data(pv)

        # database stream
        put_on_database(db, pv)

        # save to csv
        if pv.csv_output:
            save_csv(pv)

        elapsed = str(datetime.now() - pv.begin_time)
        pv.logger.info(" Completed point values processing, elapsed "
                       "time: {}".format(elapsed))

    # make a blank locations_csv file
    if blank:
        pv = PointValues()
        blank = pd.DataFrame(columns=pv.csv_cols)
        blank.to_csv('basin_locations.csv', index=False)
        print('\nCreated a blank locations_csv file:\n '
              '    basin_locations.csv\n\n'
              'Fill in either ["model xind", "model yind"] '
              'or ["latitude","longitude"] columns\n')




import numpy as np
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors
from mpl_toolkits.axes_grid1 import make_axes_locatable
import seaborn as sns
import cmocean
import matplotlib.patches as mpatches
import pandas as pd
import snowav.framework.figures

def swi(args, logger=None):
    '''
    SWI figure, with depth image on the left and volume by elevation band on
    the right.

    All other snowav figures reference this function for args description.

    Args
    ----------
    args : dict
        dictionary of values and options for figure. See CoreConfig.ini
            for more information.

            df: standard subbasin by elevation dataframe from
                snowav.database.database.collect()
                       Extended Tuolumne  Tuolumne  Cherry Creek  Eleanor
                3000      0.008              0.008         0.000    0.000
                ...
            image: 2D array of accumulated SWI
            title: figure title
            directory: figure directory name
            figs_path: base directory for saving figures
            edges: array of elevation bins in min, max, step
            plotorder: list of basins from topo.nc
            labels: dictionary of labels to use with plotorder
            lims: more figure specs from snowav.plotting.plotlims(plotorder)
            masks: masks made from topo.nc and masks()
            figsize: figure size
            dpi: figure dpi
            depthlbl: depth label
            vollbl: volume label
            dplcs: decimal places to round outputs
            barcolors: list of colors for bar plots
            xlims: xlims
            elevlbl: elevation label
            depth_clip: lower limit on image depths for plotting
            percent_min: quantile min for image
            percent_max: quantile max for image

    logger : list
        snowav logger

    '''

    # print to screen for each figure if desired
    if args['print']:
        print('swi() figure args:\n','omitting masks and image...\n')
        for name in args.keys():
            if name not in ['masks','image']:
                print(name, ': ', args[name])

    # These get used more and we'll just assign here
    masks = args['masks']
    accum = args['image']
    df = args['df']
    plotorder = args['plotorder']
    lims = args['lims']
    edges = args['edges']
    labels = args['labels']
    barcolors = args['barcolors']

    qMin,qMax = np.nanpercentile(accum,[0,args['percent_max']])
    clims = (0,qMax)
    colors1 = cmocean.cm.dense(np.linspace(0, 1, 255))
    colors2 = plt.cm.Set1_r(np.linspace(0, 1, 1))
    colors = np.vstack((colors2, colors1))
    mymap = mcolors.LinearSegmentedColormap.from_list('my_colormap', colors)

    pmask = masks[plotorder[0]]['mask']
    ixo = pmask == 0
    accum[ixo] = np.nan
    mymap.set_bad('white',1.)

    # Now set SWI-free to some color
    r = ~np.isnan(accum)
    r[r] &= accum[r] < 0.001
    accum[r] = 0

    sns.set_style('darkgrid')
    sns.set_context("notebook")

    plt.close(0)
    fig,(ax,ax1) = plt.subplots(num=0,
                                figsize=args['figsize'],
                                dpi=args['dpi'],
                                nrows=1,
                                ncols=2)

    h = ax.imshow(accum, cmap=mymap, clim=clims)

    for name in masks:
        ax.contour(masks[name]['mask'], cmap='Greys',linewidths=1)

    h.axes.get_xaxis().set_ticks([])
    h.axes.get_yaxis().set_ticks([])
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="4%", pad=0.2)
    cbar = plt.colorbar(h, cax = cax)
    cbar.set_label('[{}]'.format(args['depthlbl']))
    h.axes.set_title(args['title'])

    # Plot the bars
    for iters,name in enumerate(lims.sumorder):
        if df[plotorder[0]].sum() < 9.9:
            lbl = '{}: {} {}'.format(labels[name],str(int(df[name].sum())),
                                args['vollbl'])
        else:
            lbl = '{}: {} {}'.format(labels[name],str(np.round(df[name].sum(),
                                args['dplcs'])),args['vollbl'])

        if iters == 0:
            ax1.bar(range(0,len(edges)),df[name],color=args['barcolors'][iters],
                    edgecolor = 'k',label = lbl)

        else:
            ax1.bar(range(0,len(edges)),df[name],
                    bottom = pd.DataFrame(df[lims.sumorder[0:iters]]).sum(axis = 1).values,
                    color = barcolors[iters], edgecolor = 'k',label = lbl)

    ax1.xaxis.set_ticks(range(0,len(edges)))
    plt.tight_layout()
    ax1.set_xlim((args['xlims'][0]-0.5,args['xlims'][1]-0.5))

    edges_lbl = []
    for i in range(0,len(edges)):
        edges_lbl.append(str(int(edges[int(i)])))

    ax1.set_xticklabels(str(i) for i in edges_lbl)
    for tick in ax1.get_xticklabels():
        tick.set_rotation(30)

    ax1.set_xlim((args['xlims'][0]-0.5,args['xlims'][1]-0.5))

    ax1.set_ylabel('{}'.format(args['vollbl']))
    ax1.set_xlabel('elevation [{}]'.format(args['elevlbl']))
    ax1.yaxis.set_label_position("right")
    ax1.yaxis.tick_right()

    ylims = ax1.get_ylim()
    max = ylims[1] + ylims[1]*0.6
    ax1.set_ylim((0, max))

    fig.subplots_adjust(top=0.88)

    # If there is meaningful snow-free area, include patch and label
    if sum(sum(r)) > 1000:
        patches = [mpatches.Patch(color='grey', label='no SWI')]
        ax.legend(handles=patches, bbox_to_anchor=(lims.pbbx, 0.05),
                  loc=2, borderaxespad=0. )

    plt.tight_layout()

    # fix so that legend isn't obscured
    if len(plotorder) > 1:
        ax1.legend(loc=2,markerscale = 0.5)

    fig_name_short = 'swi_'
    fig_name = '{}{}{}.png'.format(args['figs_path'],fig_name_short,args['directory'])
    if logger is not None:
        logger.info(' saving {}'.format(fig_name))
    snowav.framework.figures.save_fig(fig, fig_name)

    return fig_name_short



from datetime import datetime
from matplotlib import pyplot as plt
import mysql.connector
import netCDF4 as nc
import numpy as np
import os
import pandas as pd
import seaborn as sns
import utm

from snowav.utils.stats import nashsutcliffe
import snowav.framework.figures


def stn_validate(rundirs, lbls, client, end_date, wy, snow_x, snow_y, stns, px,
                 py, login, figs_path, fig_name, dem, logger=None, factor=25.4,
                 nash_sut_flag=False, ncfile='snow.nc', nc_var='specific_mass',
                 index_col='date_time', elevlbl='ft', tbl='tbl_level1',
                 var='snow_water_equiv', dpi=200):
    """ SWE validation at snow pillow sites.

    Args
    ----------
    rundirs {list}: list of rundirs
    lbls {list}: stn labels
    client {str}: weather db client
    end_date {datetime}: period end date
    wy {int}: water year
    snow_x {}: netcdf x vector
    snow_y {}: netcdf y vector
    stns {list}: list of stations
    px {arr}: +- 1 array
    py {arr}: +- 1 array
    login {dict}: database login info
    figs_path {str}: base path for fig saving
    fig_name {str}: base str for fig name
    dem {arr}: dem
    logger {class}: snowav logger
    nash_sut_flag {bool}: include nash sutcliffe
    ncfile {str}: name of snow.nc
    nc_var {str}: name of nc variable
    indec_col {str}: weather db index
    dpi {int}: fig dpi
    elevlbl {str}: elevation label
    tbl {str}: weather db table
    factor {float}: conversion from mm
    logger {class}: snowav logger

    Returns
    ------
    flag {bool}: if there are errors stn_validate() figure will be removed
        from the report
    """

    logger.debug(' Beginning stn_validate()...')

    st_time = '{}-10-1 00:00:00'.format(str(wy-1))
    end_time = end_date.date().strftime("%Y-%-m-%-d")
    flag = True
    wy_start = datetime(wy-1, 10, 1)
    pr = range(0, 9)

    # container for +- 1 pixel
    stns_pixel = []
    for sta in stns:
        stns_pixel = stns_pixel + ['{}_'.format(sta) + str(s) for s in pr]

    # containers for measure and model results
    measure = pd.DataFrame(index=pd.date_range(wy_start, end_date, freq='D'),
                           columns=stns)
    model = pd.DataFrame(index=pd.date_range(wy_start, end_date, freq='D'),
                         columns=stns_pixel)

    # weather db query
    qry = ('SELECT tbl_metadata.* FROM tbl_metadata INNER JOIN tbl_stations ON '
           'tbl_metadata.id = tbl_stations.metadata_id WHERE '
           'tbl_stations.client="{}";'.format(client))

    cnx = mysql.connector.connect(user=login['user'],
                                  password=login['password'],
                                  host=login['host'],
                                  port=login['port'],
                                  database='weather_db')

    metadata = pd.read_sql(qry, cnx)

    if metadata.empty:
        flag = False
        return

    metadata.index = metadata['primary_id']
    metadata = metadata[~metadata.index.duplicated(keep='first')]

    # station results
    for iters, stn in enumerate(stns):
        cnx = mysql.connector.connect(user=login['user'],
                                      password=login['password'],
                                      host=login['host'],
                                      port=login['port'],
                                      database='weather_db')

        qry = ('SELECT weather_db.{0}.date_time, weather_db.{0}.{1} FROM '
               'weather_db.{0} WHERE weather_db.{0}.date_time between "{2}" '
               'and "{3}" AND weather_db.{0}.station_id IN ("{4}")'
               ';'.format(tbl, var, st_time, end_time, stn))

        # data = pd.read_sql(qry, cnx, index_col=bytes(bytearray(b'date_time')))
        data = pd.read_sql(qry, cnx, index_col=index_col)
        data.index.names = [index_col]
        dind = pd.date_range(st_time, end_time, freq='D')
        measure[stn] = data.reindex(dind)

    sns.set_style('darkgrid')
    sns.set_context('notebook')

    plt.close(9)

    if len(stns) <= 3:
        fig, axs = plt.subplots(num=9, figsize=(6, 6), dpi=dpi,
                                nrows=3, ncols=1)
    if (len(stns) > 3) and (len(stns) <= 6):
        fig, axs = plt.subplots(num=9, figsize=(10, 10), dpi=dpi,
                                nrows=3, ncols=2)
    if (len(stns) > 6) and (len(stns) <= 9):
        fig, axs = plt.subplots(num=9, figsize=(10, 10), dpi=dpi,
                                nrows=3, ncols=3)
    if (len(stns) > 9) and (len(stns) <= 12):
        fig, axs = plt.subplots(num=9, figsize=(10, 10), dpi=dpi,
                                nrows=4, ncols=3)
    axs = axs.flatten()

    set_x_on = 12
    if len(stns) in (2, 5, 8):
        fig.delaxes(axs[len(stns)])
        if len(stns) == 5:
            set_x_on = 3
        if len(stns) == 8:
            set_x_on = 5
    if len(stns) == 10:
        fig.delaxes(axs[10])
        fig.delaxes(axs[11])
    if len(stns) == 11:
        fig.delaxes(axs[11])

    for rname in rundirs:
        logger.debug(' Loading pixel values in '
                     '{}...'.format(rname.split('runs')[-1]))

        snowfile = os.path.join(rname, ncfile)

        if not os.path.isfile(snowfile):
            raise Exception('invalid file --> {}'.format(snowfile))

        ncf = nc.Dataset(snowfile, 'r')
        t = nc.num2date(ncf.variables['time'][0], ncf.variables['time'].units)

        # Only load the rundirs that we need
        if t.date() <= end_date.date():

            for iters, stn in enumerate(stns):
                ll = utm.from_latlon(metadata.ix[stn, 'latitude'],
                                     metadata.ix[stn, 'longitude'])
                xind = np.where(abs(snow_x - ll[0]) ==
                                min(abs(snow_x - ll[0])))[0]
                yind = np.where(abs(snow_y - ll[1]) ==
                                min(abs(snow_y - ll[1])))[0]
                elev = dem[yind, xind]

                if ((xind == 0) or (xind >= len(snow_x) - 1) or
                        (yind == 0) or (yind >= len(snow_y) - 1)):
                    logger.info('Station {} in [validate] stations is '
                                'outside of domain, exiting stn_validate() '
                                'and not generating figure'.format(stn))
                    flag = False
                    return flag

                for ix, (n, m) in enumerate(zip(px, py)):
                    land_stn = stn + '_{}'.format(str(ix))
                    swe = ncf.variables[nc_var][:, yind + m, xind + n][0][0][0]
                    model.loc[t.date(), land_stn] = swe

                    if (n == 0 and m == 0) and (rname == rundirs[-1]):
                        lbl = 'measured'
                        lblm = 'modeled'
                    else:
                        lbl = '__nolabel__'
                        lblm = 'modeled'

                    axs[iters].plot(measure[stn] / factor,
                                    'k',
                                    label=lbl)
                    axs[iters].plot(model[land_stn] / factor,
                                    'b',
                                    linewidth=0.75,
                                    label=lblm)
                    axs[iters].set_title('{} [{} {}]'.format(lbls[iters],
                                                             str(int(elev)),
                                                             elevlbl))
                    axs[iters].set_xlim((datetime(wy - 1, 10, 1), end_date))

            ncf.close()

        else:
            ncf.close()

    # Nash-Sutcliffe on pixel 0,0
    if nash_sut_flag:
        for iters, sta in enumerate(stns):
            pz_sta = sta + '_5'
            stime = datetime(wy - 1, 10, 15)
            nsv = nashsutcliffe(measure[measure.index > stime][stn].values.tolist(),
                                model[model.index > stime][pz_sta].values.tolist())
            nstr = 'Nash-Sutcliffe:\n{}'.format(str(np.round(nsv, 2)))

            nsx = 0.05
            nsy = 0.85

            if iters == 0:
                nsy = 0.675
            if nsv > 0:
                axs[iters].text(nsx, nsy, nstr, horizontalalignment='left',
                                transform=axs[iters].transAxes, fontsize=8)

    if len(stns) <= 6:
        for n in (1, 3, 5):
            axs[n].yaxis.tick_right()
        for n in (4, 5):
            for tick in axs[n].get_xticklabels():
                tick.set_rotation(30)
        for n in (0, 1, 2, 3):
            if n != set_x_on:
                axs[n].set_xticklabels('')
            else:
                for tick in axs[n].get_xticklabels():
                    tick.set_rotation(30)

    if (len(stns) > 6) and (len(stns) <= 9):
        for n in (2, 5, 8):
            axs[n].yaxis.tick_right()

        for n in (6, 7, 8):
            for tick in axs[n].get_xticklabels():
                tick.set_rotation(30)
        for n in (0, 1, 2, 3, 4, 5):
            axs[n].set_xticklabels('')
        for n in (1, 4, 7):
            axs[n].set_yticklabels('')

    if (len(stns) > 9) and (len(stns) <= 12):
        for n in (2, 5, 8, 11):
            axs[n].yaxis.tick_right()
        for n in (9, 10, 11):
            for tick in axs[n].get_xticklabels():
                tick.set_rotation(30)
        for n in (0, 1, 2, 3, 4, 5, 6, 7, 8):
            axs[n].set_xticklabels('')

    # Plot
    measure = measure.replace('[]', np.nan)
    maxm = np.nanmax(measure.max().values / factor)
    maxi = np.nanmax(model.max().values / factor)

    if maxm > maxi:
        maxswe = maxm
    else:
        maxswe = maxi

    for iters in range(0, len(stns)):
        axs[iters].set_ylim((-0.1, maxswe + maxswe * 0.05))

    axs[0].legend(['measured', 'modeled'], loc='upper left')
    axs[0].set_ylabel('SWE [in]')

    plt.suptitle('SWE Validation at Snow Pillow Sites')
    plt.subplots_adjust(top=0.92)

    fig_name = os.path.join(os.path.abspath(figs_path), fig_name)

    snowav.framework.figures.save_fig(fig, fig_name)

    if logger is not None:
        logger.info(' Saved: {}'.format(fig_name))

    return flag



import copy
from datetime import datetime, timedelta
import numpy as np
import os
import pandas as pd

from tablizer.tablizer import get_existing_records
from snowav.plotting.swi import swi
from snowav.plotting.basin_total import basin_total
from snowav.plotting.cold_content import cold_content
from snowav.plotting.compare_runs import compare_runs
from snowav.plotting.density import density
from snowav.plotting.flt_image_change import flt_image_change
from snowav.plotting.image_change import image_change
from snowav.plotting.precip_depth import precip_depth
from snowav.plotting.stn_validate import stn_validate
from snowav.plotting.write_properties import write_properties
from snowav.plotting.swe_volume import swe_volume
from snowav.plotting.inputs import inputs
from snowav.inflow.inflow import inflow
from snowav.plotting.diagnostics import diagnostics
from snowav.plotting.point_values import point_values_csv, point_values_figures
from snowav.database.database import collect
from snowav.plotting.plotlims import plotlims as plotlims
import matplotlib as mpl

if os.environ.get('DISPLAY', '') == '':
    mpl.use('Agg')


def figures(cfg, process):
    """ Set up and call snowav figures. See CoreConfig.ini and README.md for
    more on config options and use.

    swe_volume() must be called before cold_content() if you want to use
    the same ylims for each.

    Args
    ------
    cfg: config object
    process: process object

    """

    args = {'report_start': cfg.report_start.date().strftime("%Y-%-m-%-d"),
            'report_date': cfg.report_date.date().strftime("%Y-%-m-%-d"),
            'print': cfg.print_args_dict,
            'run_name': cfg.run_name,
            'start_date': cfg.start_date,
            'end_date': cfg.end_date,
            'directory': cfg.directory,
            'figs_path': cfg.figs_path,
            'edges': cfg.edges,
            'plotorder': cfg.plotorder,
            'labels': cfg.labels,
            'lims': plotlims(cfg.plotorder),
            'masks': cfg.masks,
            'figsize': cfg.figsize,
            'dpi': cfg.dpi,
            'depthlbl': cfg.depthlbl,
            'vollbl': cfg.vollbl,
            'elevlbl': cfg.elevlbl,
            'dplcs': cfg.dplcs,
            'barcolors': cfg.barcolors,
            'xlims': cfg.xlims,
            'depth_clip': cfg.depth_clip,
            'percent_min': cfg.percent_min,
            'percent_max': cfg.percent_max,
            'basins': cfg.basins,
            'wy': cfg.wy,
            'flag': False,
            'flt_flag': cfg.flt_flag}

    if cfg.flt_flag:
        args['flight_dates'] = cfg.flight_diff_dates

    fig_names = {}
    connector = cfg.connector

    ##########################################################################
    #       For each figure, collect 2D array image, by-elevation            #
    #       DataFrame, and set any figure-specific args inputs               #
    ##########################################################################
    if cfg.flt_flag:
        args['depth_factor'] = cfg.depth_factor
        args['update_file'] = cfg.update_file
        args['update_numbers'] = cfg.update_numbers
        args['flight_outputs'] = cfg.flight_outputs
        args['pre_flight_outputs'] = cfg.pre_flight_outputs
        args['connector'] = connector

        cfg.flight_diff_fig_names, cfg.flight_delta_vol_df = flt_image_change(args, cfg._logger)

        if cfg.flight_diff_fig_names == []:
            cfg.flt_flag = False

    if cfg.swi_flag:
        image = np.zeros_like(cfg.outputs['swi_z'][0])
        for n in range(cfg.ixs, cfg.ixe):
            image = image + cfg.outputs['swi_z'][n] * cfg.depth_factor

        df = collect(connector, args['plotorder'], args['basins'],
                     args['start_date'], args['end_date'], 'swi_vol',
                     args['run_name'], args['edges'], 'sum')

        args['df'] = df
        args['image'] = image
        args['title'] = 'Accumulated SWI\n{} to {}'.format(
            args['report_start'],
            args['report_date'])

        fig_names['swi'] = swi(args, cfg._logger)

    if cfg.image_change_flag:
        image = cfg.outputs['swe_z'][cfg.ixe] - cfg.outputs['swe_z'][cfg.ixs]

        start = collect(connector, args['plotorder'], args['basins'],
                        args['start_date'], args['start_date'], 'swe_vol',
                        args['run_name'], args['edges'], 'end')
        end = collect(connector, args['plotorder'], args['basins'],
                      args['start_date'], args['end_date'], 'swe_vol',
                      args['run_name'], args['edges'], 'end')

        df = end - start

        args['df'] = df
        args['image'] = image * cfg.depth_factor
        args['title'] = 'Change in SWE\n{} to {}'.format(
            args['report_start'], args['report_date'])

        fig_names['image_change'] = image_change(args, cfg._logger)

    if cfg.swe_volume_flag:
        df = collect(connector, args['plotorder'], args['basins'],
                     args['start_date'], args['end_date'], 'swe_vol',
                     args['run_name'], args['edges'], 'end')

        image = cfg.outputs['swe_z'][cfg.ixe] * cfg.depth_factor

        args['df'] = df
        args['image'] = image
        args['title'] = 'SWE {}'.format(args['report_date'])

        fig_names['swe_volume'], args['ylims'] = swe_volume(args, cfg._logger)

    if cfg.cold_content_flag:
        df = collect(connector, args['plotorder'], args['basins'],
                     args['start_date'], args['end_date'], 'swe_unavail',
                     args['run_name'], args['edges'], 'end')

        swe = cfg.outputs['swe_z'][cfg.ixe]
        image = cfg.outputs['coldcont'][cfg.ixe] * 0.000001

        args['df'] = df
        args['swe'] = swe
        args['image'] = image
        args['title'] = 'Cold Content {}'.format(args['report_date'])

        fig_names['cold_content'] = cold_content(args, cfg._logger)

    if cfg.density_flag:
        image = cfg.outputs['density'][cfg.ixe]
        args['density'] = process.density
        args['image'] = image
        args['title'] = 'Density {}'.format(args['report_date'])

        fig_names['density'] = density(args, cfg._logger)

    if cfg.basin_total_flag:
        wy_start = datetime(cfg.wy - 1, 10, 1)
        swi_summary = collect(connector, args['plotorder'], args['basins'],
                              wy_start, args['end_date'], 'swi_vol',
                              args['run_name'], 'total', 'daily')
        df_swe = collect(connector, args['plotorder'], args['basins'],
                         wy_start, args['end_date'], 'swe_vol',
                         args['run_name'], 'total', 'daily')
        df_swi = swi_summary.cumsum()

        args['swi_summary'] = df_swi
        args['swe_summary'] = df_swe
        args['forecast_flag'] = cfg.forecast_flag
        args['flt_flag'] = cfg.flt_flag

        fig_names['basin_total'] = basin_total(args, cfg._logger)

    if cfg.precip_depth_flag:
        swi_image = np.zeros_like(cfg.outputs['swi_z'][0])
        for n in range(cfg.ixs, cfg.ixe):
            swi_image = swi_image + cfg.outputs['swi_z'][n] * cfg.depth_factor

        swi_df = collect(connector, args['plotorder'], args['basins'],
                         args['start_date'], args['end_date'], 'swi_z',
                         args['run_name'], args['edges'], 'sum')
        precip_df = collect(connector, args['plotorder'], args['basins'],
                            args['start_date'], args['end_date'], 'precip_z',
                            args['run_name'], args['edges'], 'sum')
        rain_df = collect(connector, args['plotorder'], args['basins'],
                          args['start_date'], args['end_date'], 'rain_z',
                          args['run_name'], args['edges'], 'sum')

        args['swi_image'] = swi_image
        args['precip_image'] = process.precip_total * cfg.depth_factor
        args['rain_image'] = process.rain_total * cfg.depth_factor
        args['swi_df'] = swi_df
        args['precip_df'] = precip_df
        args['rain_df'] = rain_df
        args['title'] = 'Depth of SWI, Precipitation, and Rain\n{} to {}'.format(
            args['report_start'], args['report_date'])

        fig_names['precip_depth'] = precip_depth(args, cfg._logger)

    if cfg.diagnostics_flag:
        wy_start = datetime(cfg.wy - 1, 10, 1)
        precip = collect(connector, args['plotorder'], args['basins'],
                         wy_start, args['end_date'], 'precip_z',
                         args['run_name'], args['edges'], 'daily')
        precip_per = collect(connector, args['plotorder'], args['basins'],
                             args['start_date'], args['end_date'], 'precip_z',
                             args['run_name'], args['edges'], 'daily')
        swe = collect(connector, args['plotorder'], args['basins'],
                      wy_start, args['end_date'], 'swe_z',
                      args['run_name'], args['edges'], 'daily')
        swe_per = collect(connector, args['plotorder'], args['basins'],
                          args['start_date'], args['end_date'], 'swe_z',
                          args['run_name'], args['edges'], 'daily')
        rho = collect(connector, args['plotorder'], args['basins'],
                      wy_start, args['end_date'], 'density',
                      args['run_name'], args['edges'], 'daily')
        rho_per = collect(connector, args['plotorder'], args['basins'],
                          args['start_date'], args['end_date'], 'density',
                          args['run_name'], args['edges'], 'daily')
        snow_line = collect(connector, args['plotorder'], args['basins'],
                            wy_start, args['end_date'], 'snow_line',
                            args['run_name'], args['edges'], 'daily')

        snow_line_per = collect(connector, args['plotorder'], args['basins'],
                                args['start_date'], args['end_date'], 'snow_line',
                                args['run_name'], args['edges'], 'daily')
        evap_z = collect(connector, args['plotorder'], args['basins'],
                         wy_start, args['end_date'], 'evap_z',
                         args['run_name'], args['edges'], 'daily')
        evap_z_per = collect(connector, args['plotorder'], args['basins'],
                             args['start_date'], args['end_date'], 'evap_z',
                             args['run_name'], args['edges'], 'daily')

        snow_line_per = snow_line_per.fillna(0)
        first_row = snow_line_per.iloc[[0]].values[0]
        snow_line_per = snow_line_per.apply(lambda row: row - first_row, axis=1)
        args['snow_line'] = snow_line
        args['snow_line_per'] = snow_line_per

        swe = swe.fillna(0)
        swe_per = swe_per.fillna(0)
        first_row = swe_per.iloc[[0]].values[0]
        swe_per = swe_per.apply(lambda row: row - first_row, axis=1)

        evap_z = evap_z.fillna(0)
        evap_z_per = evap_z_per.fillna(0)
        first_row = evap_z_per.iloc[[0]].values[0]
        evap_z_per = evap_z_per.apply(lambda row: row - first_row, axis=1)

        rho = rho.fillna(0)
        rho_per = rho_per.fillna(0)
        first_row = rho_per.iloc[[0]].values[0]
        rho_per = rho_per.apply(lambda row: row - first_row, axis=1)

        precip = precip.fillna(0)
        precip_per = precip_per.fillna(0)
        precip = precip.cumsum()
        precip_per = precip_per.cumsum()
        first_row = precip_per.iloc[[0]].values[0]
        precip_per = precip_per.apply(lambda row: row - first_row, axis=1)

        if cfg.diag_basins is None:
            args['dbasins'] = copy.deepcopy(cfg.plotorder)
        else:
            args['dbasins'] = cfg.diag_basins

        args['precip'] = precip
        args['precip_per'] = precip_per
        args['swe'] = swe
        args['swe_per'] = swe_per
        args['evap_z'] = evap_z
        args['evap_z_per'] = evap_z_per
        args['density'] = rho
        args['density_per'] = rho_per
        args['elevlbl'] = cfg.elevlbl

        diagnostics(args, cfg._logger)

    if cfg.stn_validate_flag:
        px = (1, 1, 1, 0, 0, 0, -1, -1, -1)
        py = (1, 0, -1, 1, 0, -1, 1, 0, -1)
        et = args['end_date'].date().strftime("%Y%-m%-d")
        basin = cfg.plotorder[0].lower().split(" ")[0]
        name = '{}_validation_{}.png'.format(basin, et)
        login = {'user': cfg.wxdb_user,
                 'password': cfg.wxdb_password,
                 'host': cfg.wxdb_host,
                 'port': cfg.wxdb_port}

        # assign fig name to cfg for use in report.py
        cfg.assign_vars({'stn_validate_fig_name': name})

        flag = stn_validate(cfg.all_dirs, cfg.val_lbls, cfg.val_client,
                            args['end_date'], args['wy'], cfg.snow_x,
                            cfg.snow_y, cfg.val_stns, px, py, login,
                            args['figs_path'], name, cfg.dem,
                            logger=cfg._logger, elevlbl=args['elevlbl'],
                            nash_sut_flag=cfg.nash_sut_flag)

        if not flag:
            cfg.stn_validate_flag = False

    else:
        # assign fig name to cfg for use in report.py
        cfg.assign_vars({'stn_validate_fig_name': ''})

    if cfg.point_values:
        cfg._logger.debug(" Beginning point values processing for "
                          "{}".format(cfg.point_values_csv))

        flag = True
        xy = (cfg.snow_x, cfg.snow_y)
        headings = ['name', 'latitude', 'longitude', cfg.point_values_heading]
        end_date_str = cfg.end_date.date().strftime("%Y-%m-%d")
        course_date = cfg.point_values_date.date().strftime('%Y-%m-%d')
        basin_name = cfg.plotorder[0].split(" ")[0].lower()
        pv_date = cfg.point_values_date
        nsubplots = (cfg.point_values_settings[3] *
                     cfg.point_values_settings[4] - 1) - 1

        while flag:
            df = pd.read_csv(cfg.point_values_csv)

            for head in headings:
                if head not in df.columns.tolist():
                    cfg._logger.warn(' Required csv column "{}" not found, '
                                     'exiting point values'.format(head))
                    if head == cfg.point_values_heading:
                        cfg._logger.warning(' User specified [validate] '
                                            'point_values_heading: {} not '
                                            'found'.format(head))
                    flag = False

            if pv_date is None:
                cfg._logger.info(' Value in [validate] point_values_date '
                                 'being assigned to {}'.format(end_date_str))
                pv_date = cfg.end_date

            if pv_date < cfg.start_date or pv_date > cfg.end_date:
                cfg._logger.info(' Value in [validate] point_values_date '
                                 'outside of range in [run] start_date - '
                                 'end_date, point_values_date being assigned '
                                 'to: {}'.format(end_date_str))
                idx = -1
            else:
                x = np.abs([date - pv_date for date in cfg.outputs['dates']])
                idx = x.argmin(0)

            model_date = cfg.outputs['dates'][idx].date().strftime('%Y-%m-%d')

            for value in cfg.point_values_properties:
                filename = '{}_pixel_{}_{}.csv'.format(basin_name, value,
                                                       end_date_str)
                csv_name = os.path.abspath(os.path.join(cfg.figs_path,
                                                        filename))

                if len(df.name.unique()) > nsubplots:
                    cfg._logger.warn(' Number of subplots in '
                                     'point_values() may not fit well with '
                                     'given settings, consider changing '
                                     'nrows and/or ncols in [validate] '
                                     'point_values_settings')
                    flag = False

                if value == 'swe_z':
                    factor = cfg.depth_factor
                elif value == 'depth':
                    factor = 39.37
                else:
                    factor = 1

                array = cfg.outputs[value][idx] * factor

                df_res = point_values_csv(array, value, df, xy, csv_name,
                                          model_date, cfg.plotorder[0],
                                          cfg.point_values_heading,
                                          cfg._logger)

                if cfg.point_values_flag:
                    head = cfg.point_values_heading
                    if head in df.columns.tolist():
                        point_values_figures(array, value, df_res, cfg.dem,
                                             cfg.figs_path, cfg.veg_type,
                                             model_date, course_date,
                                             cfg.point_values_settings,
                                             cfg.pixel, head, cfg._logger)
                    else:
                        cfg._logger.warn(' [validate] point_values_heading: '
                                         '{} not in csv, skipping figures '
                                         ''.format(cfg.point_values_heading))
                        cfg.point_values_flag = False

            # if everything is successful, set to False at the end
            flag = False

    if cfg.compare_runs_flag:
        args['variables'] = ['swe_vol', 'swi_vol']

        if cfg.flt_flag:
            args['flag'] = True
        else:
            args['flag'] = False

        dict = {}
        for var in args['variables']:
            dict[var] = {}
            for wy, run in zip(cfg.compare_run_wys, cfg.compare_run_names):
                wy_start = datetime(wy - 1, 10, 1)
                df = collect(connector, args['plotorder'][0], args['basins'],
                             wy_start, args['end_date'], var, run, 'total', 'daily')

                if wy != cfg.wy:
                    adj = cfg.wy - wy
                    df.index = df.index + timedelta(days=365 * adj)

                if var == 'swi_vol':
                    df = df.cumsum()

                dict[var][run] = df

        args['dict'] = dict

        compare_runs(args, cfg._logger)

    if cfg.inflow_flag:
        wy_start = datetime(cfg.wy - 1, 10, 1)
        swi_summary = collect(connector, args['plotorder'], args['basins'],
                              wy_start, args['end_date'], 'swi_vol',
                              args['run_name'], 'total', 'daily')
        df_swi = swi_summary.cumsum()

        args['swi_summary'] = df_swi

        if cfg.inflow_data is None:
            raw = pd.read_csv(cfg.summary_csv, skiprows=1,
                              parse_dates=[0], index_col=0)
            args['inflow_summary'] = raw[cfg.basin_headings]

        else:
            args['inflow_summary'] = pd.read_csv(cfg.summary_csv,
                                                 parse_dates=[0], index_col=0)

        args['inflow_headings'] = cfg.inflow_headings
        args['basin_headings'] = cfg.basin_headings

        inflow(args, cfg._logger)

    if cfg.write_properties is not None:
        write_properties(args['end_date'], cfg.connector, args['plotorder'],
                         args['basins'], datetime(cfg.wy - 1, 10, 1),
                         args['run_name'], args['figs_path'],
                         cfg.write_properties, vollbl=args['vollbl'],
                         logger=cfg._logger)

    if cfg.inputs_fig_flag:

        if cfg.mysql is not None:
            dbs = 'sql'
        else:
            dbs = 'sqlite'

        df = get_existing_records(connector, dbs)
        df = df.set_index('date_time')
        df.sort_index(inplace=True)

        ivalue = {}
        p = []

        for var in cfg.plots_inputs_variables:
            ivalue[var] = {}

            for basin in cfg.inputs_basins:
                bid = args['basins'][basin]['basin_id']
                ivalue[var][basin] = {}

                for func in cfg.inputs_methods:

                    if 'percentile' in func:
                        nfunc = '{}_{}'.format(func, str(cfg.inputs_percentiles[0]))

                        if ((var == cfg.plots_inputs_variables[0]) and
                                (basin == cfg.inputs_basins[0])):
                            p.append(nfunc)

                        ivalue[var][basin][nfunc] = df[(df['function'] == nfunc) &
                                                       (df['variable'] == var) &
                                                       (df['basin_id'] == int(bid)) &
                                                       (df['run_name'] == args['run_name'])]

                        nfunc = '{}_{}'.format(func, str(cfg.inputs_percentiles[1]))

                        if ((var == cfg.plots_inputs_variables[0]) and
                                (basin == cfg.inputs_basins[0])):
                            p.append(nfunc)

                        ivalue[var][basin][nfunc] = df[(df['function'] == nfunc) &
                                                       (df['variable'] == var) &
                                                       (df['basin_id'] == int(bid)) &
                                                       (df['run_name'] == args['run_name'])]
                    else:
                        ivalue[var][basin][func] = df[(df['function'] == func) &
                                                      (df['variable'] == var) &
                                                      (df['basin_id'] == int(bid)) &
                                                      (df['run_name'] == args['run_name'])]

                        if ((var == cfg.plots_inputs_variables[0]) and
                                (basin == cfg.inputs_basins[0])):
                            p.append(func)

        args['inputs'] = ivalue
        args['inputs_methods'] = p
        args['var_list'] = cfg.plots_inputs_variables
        args['inputs_basins'] = cfg.inputs_basins

        inputs(args, cfg._logger)

    if cfg.forecast_flag:
        print('Forecast figures in progress...')

    cfg.fig_names = fig_names


def save_fig(fig, paths):
    '''
    Args
    ----------
    fig : object
        matplotlib figure object
    paths : list
        list of paths to save figure

    '''

    if type(paths) != list:
        paths = [paths]

    for path in paths:
        fig.savefig(path)



import calendar
import coloredlogs
from copy import deepcopy
from datetime import timedelta, datetime
import logging
import numpy as np
import os
import netCDF4 as nc

from inicheck.tools import get_user_config, check_config, cast_all_variables
from inicheck.output import generate_config, print_config_report
from inicheck.config import MasterConfig
import snowav
from snowav.utils.utilities import masks, get_snowav_path
from snowav.utils.get_topo_stats import get_topo_stats
from snowav.framework.outputs import outputs
from snowav.database.database import connect
from snowav.database.models import AwsmInputsOutputs
from snowav.utils.wyhr import handle_year_stradling, calculate_date_from_wyhr


class UserConfig(object):
    """ Config file class.

    Args
    ------
    config_file: file path
    awsm: awsm object
    end_date: string
    """

    def __init__(self, config_file, awsm=None, end_date=None):

        print('Reading {} and loading files...'.format(config_file))

        self.config_file = config_file
        snowav_mcfg = MasterConfig(modules='snowav')
        ucfg = get_user_config(self.config_file, mcfg=snowav_mcfg)
        ucfg.apply_recipes()
        ucfg = cast_all_variables(ucfg, ucfg.mcfg)
        self.snowav_path = get_snowav_path()

        warnings, errors = check_config(ucfg)
        if errors != [] or warnings != []:
            print_config_report(warnings, errors)

        self.tmp_log = []
        self.tmp_err = []
        self.tmp_warn = []
        self.proc_time_start = datetime.now()

        ####################################################
        #            snowav                                #
        ####################################################
        self.loglevel = ucfg.cfg['snowav']['log_level'].upper()
        self.log_to_file = ucfg.cfg['snowav']['log_to_file']
        self.save_path = ucfg.cfg['snowav']['save_path']
        self.units = ucfg.cfg['snowav']['units']
        self.elev_bins = ucfg.cfg['snowav']['elev_bins']
        self.directory = ucfg.cfg['snowav']['directory']
        self.dempath = ucfg.cfg['snowav']['dempath']
        self.run_name = ucfg.cfg['snowav']['run_name']
        self.plotorder = ucfg.cfg['snowav']['masks']
        self.plotlabels = ucfg.cfg['snowav']['plotlabels']

        ####################################################
        #           run                                    #
        ####################################################
        self.dplcs = ucfg.cfg['run']['decimals']
        self.start_date = ucfg.cfg['run']['start_date']
        self.end_date = ucfg.cfg['run']['end_date']

        if end_date is not None:
            self.end_date = end_date
            self.tmp_log.append(' Overriding config end_date with '
                                '{} given with snowav call'.format(end_date))

            if self.end_date <= self.start_date:
                raise Exception('end_date {} earlier than start_date {}'.format(
                    self.end_date, self.start_date))

        if self.start_date is not None and self.end_date is not None:
            self.start_date = self.start_date
            self.end_date = self.end_date

            if self.start_date >= self.end_date:
                self.tmp_log.append(' Error: [run] start_date >= end_date')
                raise Exception('[run] start_date >= [run] end_date')

        else:
            self.tmp_log.append(' [run] start_date and/or end_date was not '
                                'defined in config file, will be assigned '
                                'by available dates in directory')

        self.all_subdirs = ucfg.cfg['run']['all_subdirs']

        if (ucfg.cfg['run']['directory'] is None) and (awsm is not None):
            if self.all_subdirs is True:
                self.run_dirs = ([awsm.pathr + s for s in
                                  os.listdir(awsm.pathr)
                                  if (os.path.isdir(awsm.pathr + s))])
            else:
                self.run_dirs = awsm.pathr
                if type(self.run_dirs) != list:
                    self.run_dirs = [self.run_dirs]

        else:
            directory = ucfg.cfg['run']['directory']

            if len(directory) == 1:
                directory = directory[0]

            if self.all_subdirs is True:
                self.run_dirs = ([directory + s for s in os.listdir(directory)
                                  if (os.path.isdir(directory + s))])
            else:
                self.run_dirs = ucfg.cfg['run']['directory']
                if type(self.run_dirs) != list:
                    self.run_dirs = [self.run_dirs]

        self.run_dirs.sort()

        ####################################################
        #         database
        ####################################################
        self.mysql = ucfg.cfg['database']['mysql']
        self.db_user = ucfg.cfg['database']['user']
        self.db_password = ucfg.cfg['database']['password']
        self.db_host = ucfg.cfg['database']['host']
        self.db_port = ucfg.cfg['database']['port']
        self.db_convert = ucfg.cfg['database']['convert_ws']
        self.add_basins = ucfg.cfg['database']['add_basins']
        self.db_overwrite = ucfg.cfg['database']['overwrite']
        self.properties = ucfg.cfg['database']['properties']
        self.sqlite = ucfg.cfg['database']['sqlite']

        base_bands = ['swi_z', 'evap_z', 'swe_z', 'depth', 'density',
                      'coldcont', 'precip_z']

        for band in base_bands:
            if band not in self.properties:
                self.tmp_log.append(' WARNING! Config option [database] '
                                    'properties does not contain '
                                    '{}'.format(band))

        if ((self.mysql is not None) and
                ((self.db_user is None) or
                 (self.db_password is None) or
                 (self.db_host is None) or
                 (self.db_port is None))):
            raise Exception('If using config option [database] mysql, must '
                            'also supply user, password, host, and port')

        if self.sqlite is not None:
            if not os.path.isdir(os.path.dirname(self.sqlite)):
                raise Exception('{} does not contain a valid base '
                                'path'.format(self.sqlite))
            self.sqlite = 'sqlite:///' + self.sqlite

            if self.mysql is not None:
                raise Exception('Config option [database] section contains '
                                'both "mysql" and "sqlite" entries, pick one.')

        ####################################################
        #           validate                               #
        ####################################################
        self.val_stns = ucfg.cfg['validate']['stations']
        self.val_lbls = ucfg.cfg['validate']['labels']
        self.val_client = ucfg.cfg['validate']['client']
        self.pre_val_stns = ucfg.cfg['validate']['pre_stations']
        self.pre_val_lbls = ucfg.cfg['validate']['pre_labels']
        self.wxdb_user = ucfg.cfg['validate']['user']
        self.wxdb_password = ucfg.cfg['validate']['password']
        self.wxdb_host = ucfg.cfg['validate']['host']
        self.wxdb_port = ucfg.cfg['validate']['port']
        self.point_values = ucfg.cfg['validate']['point_values']
        self.point_values_csv = ucfg.cfg['validate']['point_values_csv']
        self.point_values_date = ucfg.cfg['validate']['point_values_date']
        self.point_values_properties = ucfg.cfg['validate']['point_values_properties']
        self.point_values_heading = ucfg.cfg['validate']['point_values_heading']
        self.point_values_settings = ucfg.cfg['validate']['point_values_settings']

        for n in range(0, 10):
            self.point_values_settings[n] = int(self.point_values_settings[n])

        if self.point_values and self.point_values_csv is None:
            self.point_values = False
            self.tmp_log.append(' Config option [validate] point_values_csv '
                                'was not supplied, point_values being set '
                                'to False')

        if self.point_values and self.point_values_date is None:
            self.point_values = False
            self.tmp_log.append(' Config option [validate] point_values_date '
                                'was not supplied, point_values being set '
                                'to False')

        ####################################################
        #           diagnostics                            #
        ####################################################
        self.diagnostics_flag = ucfg.cfg['diagnostics']['diagnostics']
        self.diag_basins = ucfg.cfg['diagnostics']['basins']
        self.diag_limit = ucfg.cfg['diagnostics']['limit']

        if self.diagnostics_flag:
            if self.diag_basins is not None:
                for basin in self.diag_basins:
                    if basin not in self.plotorder:
                        self.tmp_log.append(' Config [diagnostics] basin: "{}"'
                                            ' does not match [snowav] masks: '
                                            '"{}", diagnostics set to '
                                            'False'.format(basin,
                                                           self.plotorder))
                        self.diagnostics_flag = False

        if 'snow_line' not in self.properties and self.diagnostics_flag:
            self.diagnostics_flag = False
            self.tmp_log.append(' Required properties in [database] properties'
                                ' for [diagnostics] does not exist, setting '
                                'diagnostics: False')

        self.inputs_flag = ucfg.cfg['diagnostics']['inputs_table']
        self.inputs_variables = ucfg.cfg['diagnostics']['inputs_variables']
        self.inputs_percentiles = ucfg.cfg['diagnostics']['inputs_percentiles']
        self.inputs_methods = ucfg.cfg['diagnostics']['inputs_methods']
        self.inputs_basins = ucfg.cfg['diagnostics']['inputs_basins']

        if self.inputs_basins is not None and self.plotorder is not None:
            for basin in self.inputs_basins:
                if basin not in self.plotorder:
                    self.tmp_log.append(' Config option [diagnostics] '
                                        'inputs_basins: {} does not match what '
                                        'was supplied in [snowav] masks: {}, '
                                        'inputs set to '
                                        'False'.format(basin, self.plotorder))
                    self.inputs_flag = False

        if self.inputs_flag:
            s = [x + ', ' for x in self.inputs_variables]
            self.tmp_log.append(' Using variables {} for inputs '
                                'summary'.format(''.join(s)))

            s = [x + ', ' for x in self.inputs_methods]
            self.tmp_log.append(' Using methods {} for inputs '
                                'summary'.format(''.join(s)))

        ####################################################
        #          plots                                   #
        ####################################################
        self.dpi = ucfg.cfg['plots']['dpi']
        self.depth_clip = ucfg.cfg['plots']['depth_clip']
        self.percent_min = ucfg.cfg['plots']['percent_min']
        self.percent_max = ucfg.cfg['plots']['percent_max']
        self.subs_fig = ucfg.cfg['plots']['subs_fig']
        self.density_flag = ucfg.cfg['plots']['density']
        self.swi_flag = ucfg.cfg['plots']['swi']
        self.current_image_flag = ucfg.cfg['plots']['current_image']
        self.image_change_flag = ucfg.cfg['plots']['image_change']
        self.cold_content_flag = ucfg.cfg['plots']['cold_content']
        self.swe_volume_flag = ucfg.cfg['plots']['swe_volume']
        self.basin_total_flag = ucfg.cfg['plots']['basin_total']
        self.stn_validate_flag = ucfg.cfg['plots']['stn_validate']
        self.nash_sut_flag = ucfg.cfg['plots']['disp_nash_sut']
        self.stns_file = ucfg.cfg['plots']['stns_file']
        self.precip_validate_flag = ucfg.cfg['plots']['precip_validate']
        self.inputs_fig_flag = ucfg.cfg['plots']['inputs']
        self.plots_inputs_variables = ucfg.cfg['plots']['inputs_variables']
        self.compare_runs_flag = ucfg.cfg['plots']['compare_runs']
        self.compare_run_names = ucfg.cfg['plots']['compare_run_names']
        self.compare_run_labels = ucfg.cfg['plots']['compare_run_labels']
        self.compare_run_wys = ucfg.cfg['plots']['compare_run_wys']
        self.precip_depth_flag = ucfg.cfg['plots']['precip_depth']
        self.basin_detail_flag = ucfg.cfg['plots']['basin_detail']
        self.update_file = ucfg.cfg['plots']['update_file']
        self.print_args_dict = ucfg.cfg['plots']['print_args_dict']
        self.figsize = (ucfg.cfg['plots']['fig_length'],
                        ucfg.cfg['plots']['fig_height'])
        self.write_properties = ucfg.cfg['plots']['write_properties']
        self.point_values_flag = ucfg.cfg['plots']['point_values']

        if (self.write_properties is not None and
                type(self.write_properties) != list):
            self.write_properties = [self.write_properties]

        numbers = ucfg.cfg['plots']['update_numbers']

        if numbers is not None:
            if type(numbers) != list:
                numbers = [numbers]
            self.update_numbers = [x - 1 for x in numbers]
        else:
            self.update_numbers = None

        if (self.compare_runs_flag and ((self.compare_run_names is None) or
            (self.compare_run_labels is None) or self.compare_run_wys is None)):
            self.tmp_log.append(' Config option [plots] compare_runs set to '
                                'True, but one of compare_run_names, '
                                'compare_run_labels, or compare_run_wys is '
                                'empty, setting compare_runs to False')
            self.compare_runs_flag = False

        if (self.compare_runs_flag and
                (len(self.compare_run_names) != len(self.compare_run_labels))):
            self.tmp_log.append(' Config option [plots] compare_runs set to '
                                'True, must supply equal length '
                                'compare_run_names and compare_run_labels, '
                                'resetting compare_runs to False')
            self.compare_runs_flag = False

        if self.update_file is not None:
            self.flt_flag = True
        else:
            self.flt_flag = False

        if (self.precip_validate_flag and ((self.val_client is None) or
            (self.pre_val_stns is None) or (self.pre_val_lbls is None))):
            self.tmp_log.append(' Config option [plots] precip_validate is '
                                'being set to False')

            self.precip_validate_flag = False

        if (self.stn_validate_flag and (self.val_client is None) or
                (self.val_stns is None) or (self.val_lbls is None) or
                (self.wxdb_user is None) or (self.wxdb_password is None)):
            self.tmp_log.append(' Config option [plots] stn_validate is being '
                                'set to False')

            self.stn_validate_flag = False

        if len(self.point_values_settings) != 14:
            self.tmp_log.append(' Expected [validate] point_values_settings '
                                'to have 14 values, point_values set to False')
            self.point_values_flag = False

        for var in self.plots_inputs_variables:
            if var not in self.inputs_variables:
                self.plots_inputs_variables.remove(var)
                self.tmp_log.append(' Config option [plots] inputs_variables '
                                    'value {} not present in [diagnostics] '
                                    'inputs_variables, being '
                                    'removed'.format(var))

        ####################################################
        #          report                                  #
        ####################################################
        self.report_flag = ucfg.cfg['report']['report']
        self.print_latex = ucfg.cfg['report']['print_latex']
        self.report_name = ucfg.cfg['report']['file']
        self.rep_title = ucfg.cfg['report']['title']
        self.rep_path = ucfg.cfg['report']['save_path']
        self.env_path = ucfg.cfg['report']['env_path']
        self.templ_path = ucfg.cfg['report']['templ_path']
        self.tex_file = ucfg.cfg['report']['tex_file']
        self.summary_file = ucfg.cfg['report']['summary']
        self.figs_tpl_path = ucfg.cfg['report']['figs_tpl_path']
        self.flight_figs = ucfg.cfg['report']['flight_figs']
        self.tables = ucfg.cfg['report']['tables']
        self.report_diagnostics = ucfg.cfg['report']['diagnostics']
        self.report_diagnostics_day = ucfg.cfg['report']['diagnostics_day']
        self.rep_dplcs = ucfg.cfg['report']['decimals']

        if (self.report_diagnostics and
                (not self.inputs_fig_flag or not self.diagnostics_flag)):
            self.tmp_log.append(" [report] diagnostics: True, but must also "
                                "have [plots] inputs: True and [diagnostics] "
                                "diagnostics: True, setting to False")
            self.report_diagnostics = False

        if self.report_diagnostics and self.report_diagnostics_day[0] != 'any':

            if (calendar.day_name[datetime.now().weekday()] not in
                    self.report_diagnostics_day):
                self.report_diagnostics = False
                self.tmp_log.append(" Per [report] diagnostics_day: {}, "
                                    "setting diagnostics: "
                                    "False".format(self.report_diagnostics_day))

        self.rep_swi_flag = ucfg.cfg['report']['swi']
        if not self.swi_flag:
            self.rep_swi_flag = False

        self.rep_image_change_flag = ucfg.cfg['report']['image_change']
        if not self.image_change_flag:
            self.rep_image_change_flag = False

        self.rep_cold_content_flag = ucfg.cfg['report']['cold_content']
        if not self.cold_content_flag:
            self.rep_cold_content_flag = False

        self.rep_swe_volume_flag = ucfg.cfg['report']['swe_volume']
        if not self.swe_volume_flag:
            self.rep_swe_volume_flag = False

        self.rep_basin_total_flag = ucfg.cfg['report']['basin_total']
        if not self.basin_total_flag:
            self.rep_basin_total_flag = False

        self.rep_stn_validate_flag = ucfg.cfg['report']['stn_validate']
        if not self.stn_validate_flag:
            self.rep_stn_validate_flag = False

        self.rep_compare_runs_flag = ucfg.cfg['report']['compare_runs']
        if not self.compare_runs_flag:
            self.rep_compare_runs_flag = False

        self.rep_precip_depth_flag = ucfg.cfg['report']['precip_depth']
        if not self.precip_depth_flag:
            self.rep_precip_depth_flag = False

        # check paths to see if they need default snowav path
        if self.env_path is None:
            self.env_path = os.path.abspath(
                os.path.join(snowav.__path__[0],
                             "report/template/section_text"))
        if self.templ_path is None:
            self.templ_path = os.path.abspath(
                os.path.join(snowav.__path__[0],
                             "report/template"))
        if self.summary_file is None:
            self.summary_file = os.path.abspath(
                os.path.join(snowav.__path__[0],
                             "report/template/section_text/report_summary.txt"))
        if self.tex_file is None:
            self.tex_file = os.path.abspath(
                os.path.join(snowav.__path__[0],
                             "report/template/snowav_report.text"))
        if self.figs_tpl_path is None:
            self.figs_tpl_path = os.path.abspath(
                os.path.join(snowav.__path__[0],
                             "report/figs"))

        ####################################################
        #           forecast                               #
        ####################################################
        self.forecast_flag = ucfg.cfg['forecast']['report']

        if self.forecast_flag:
            self.for_start_date = ucfg.cfg['forecast']['start_date']
            self.for_end_date = ucfg.cfg['forecast']['end_date']
            self.for_run_name = ucfg.cfg['forecast']['run_name']

            if self.for_start_date >= self.for_end_date:
                self.tmp_log.append(' Error: config option [forecast] '
                                    'start_date > end_date')
                raise Exception('Config option [forecast] start_date >'
                                'end_date')

            self.for_run_dir = ([ucfg.cfg['forecast']['run_dir'] + s for s in
                                 os.listdir(ucfg.cfg['forecast']['run_dir'])
                                 if (os.path.isdir(ucfg.cfg['forecast']['run_dir'] + s))])

            self.for_run_dir.sort()

        ####################################################
        #           query                                  #
        ####################################################
        self.query_flag = ucfg.cfg['query']['query']
        self.q_basins = ucfg.cfg['query']['basins']
        self.q_value = ucfg.cfg['query']['value']
        self.q_run_name = ucfg.cfg['query']['run_name']
        self.q_print_all_runs = ucfg.cfg['query']['print_all_runs']
        self.q_start_date = ucfg.cfg['query']['start_date']
        self.q_end_date = ucfg.cfg['query']['end_date']
        self.q_total = ucfg.cfg['query']['total']
        self.q_output = ucfg.cfg['query']['output']
        self.q_csv_base_path = ucfg.cfg['query']['csv_base_path']
        self.q_database = ucfg.cfg['query']['database']

        ####################################################
        #           inflow                                 #
        ####################################################
        self.inflow_flag = ucfg.cfg['inflow']['inflow']
        self.inflow_data = ucfg.cfg['inflow']['inflow_data']
        self.summary_csv = ucfg.cfg['inflow']['summary_csv']
        self.inflow_headings = ucfg.cfg['inflow']['inflow_headings']
        self.basin_headings = ucfg.cfg['inflow']['basin_headings']
        self.sheet_name = ucfg.cfg['inflow']['sheet_name']
        self.skiprows = ucfg.cfg['inflow']['skiprows']
        self.overwrite = ucfg.cfg['inflow']['overwrite']
        self.file_base = ucfg.cfg['inflow']['file_base']
        self.date_idx = ucfg.cfg['inflow']['date_idx']
        self.convert = ucfg.cfg['inflow']['convert']

        self.ucfg = ucfg

    def assign_vars(self, kwargs):
        """ Assign attributes.

        Args
        ------
        kwargs {dict}: {item, value}
        """

        if not isinstance(kwargs, dict):
            raise TypeError("kwargs must be dict")

        for item, v in kwargs.items():
            setattr(self, item, v)

    def parse(self, external_logger=None):
        """ Parse config options. """

        self.snowav_version = snowav.__version__
        self.cclimit = -5 * 1000 * 1000

        self.barcolors = ['xkcd:cobalt',
                          'xkcd:mustard green',
                          'xkcd:lichen',
                          'xkcd:pale green',
                          'xkcd:blue green',
                          'xkcd:bluish purple',
                          'xkcd:lightish purple',
                          'xkcd:deep magenta',
                          'xkcd:burgundy',
                          'red']

        out = masks(self.dempath, self.db_convert, plotorder=self.plotorder,
                    plotlabels=self.plotlabels)

        self.dem = out['dem']
        self.veg_type = out['veg_type']
        self.masks = out['masks']
        self.nrows = out['nrows']
        self.ncols = out['ncols']
        self.plotorder = out['plotorder']
        self.labels = out['labels']

        for log in out['logger']:
            self.tmp_log.append(log)

        # Establish database connection
        self.basins, cnx, out = connect(sqlite=self.sqlite, sql=self.mysql,
                                        plotorder=self.plotorder, user=self.db_user,
                                        password=self.db_password, host=self.db_host,
                                        port=self.db_port, convert=self.db_convert,
                                        add=self.add_basins)
        self.connector = cnx

        for log in out:
            self.tmp_log.append(log)

        if self.loglevel == 'DEBUG':
            for basin in self.basins:
                self.tmp_log.append(' {}: {}'.format(basin, self.basins[basin]))

        # Check snow.nc file location, get topo stats and water year
        sfile = os.path.join(self.run_dirs[0], 'snow.nc')

        if os.path.isfile(sfile):
            topo = get_topo_stats(sfile)
            self.snow_x = topo['x']
            self.snow_y = topo['y']
            self.pixel = int(topo['dv'])

            ncf = nc.Dataset(sfile)
            t = nc.num2date(ncf.variables['time'][0], ncf.variables['time'].units)
            ncf.close()
            self.wy = handle_year_stradling(t) + 1

        else:
            print('\nGiven config options, expecting to find:\n {}\nto load topo '
                  'stats but is not a valid file\nCheck config [run] options, see '
                  'CoreConfig.ini for details\n'.format(sfile))
            raise Exception('{} not a valid file'.format(sfile))

        # make the bins
        edges = np.arange(self.elev_bins[0],
                          self.elev_bins[1] + self.elev_bins[2],
                          self.elev_bins[2])

        # use for definition
        self.edges = np.arange(self.elev_bins[0] - self.elev_bins[2],
                               self.elev_bins[1],
                               self.elev_bins[2])

        v = self.properties
        if self.inputs_flag:
            v += self.inputs_variables

        # get variables that will be used in processing
        self.variables = AwsmInputsOutputs()
        self.variables.make_variables(v,
                                      self.edges,
                                      self.masks.keys())

        if self.units == 'TAF':
            self.conversion_factor = ((self.pixel ** 2) * 0.000000810713194 * 0.001)
            self.depth_factor = 0.03937
            self.dem = self.dem * 3.28
            self.depthlbl = 'in'
            self.vollbl = self.units
            self.elevlbl = 'ft'

            if max(self.edges) < 5000:
                self.tmp_log.append(" WARNING! Config options [snowav] units: TAF "
                                    "and elev_bins: {} may not match! Consider changing elev_bins "
                                    "values".format(self.elev_bins))

        if self.units == "SI":
            self.conversion_factor = ((self.pixel ** 2) * 0.000000810713194) * 1233.48 / 1e9
            self.depth_factor = 0.01
            self.depthlbl = 'cm'
            self.vollbl = 'M$M^3$'
            self.elevlbl = 'm'

            if max(self.edges) > 5000:
                self.tmp_log.append(" WARNING! Config options [snowav] units: SI "
                                    "and elev_bins: {} may not match! Consider changing elev_bins "
                                    "values".format(self.elev_bins))

        self.ixd = np.digitize(self.dem, edges)
        self.xlims = (0, len(edges))

        if self.loglevel == 'DEBUG' and self.log_to_file is not True:
            print('Reading files in {}...'.format(self.run_dirs[0].split('runs')[0]))

        results = outputs(self.run_dirs, self.wy, self.properties,
                          self.start_date, self.end_date, None, self.loglevel)

        out = results['outputs']
        all_dirs = results['dirs']
        dirs = results['run_dirs']
        rdict = results['rdict']
        log = results['log']

        # If there was an error parsing files catch and log it
        if out == [] and all_dirs == [] and 'not a valid file' in log[-1]:

            self.tmp_log.append(log[-1])
            if self.start_date is not None and self.end_date is not None:
                ext_shr = (self.directory +
                           '_' +
                           self.start_date.date().strftime("%Y%m%d") +
                           '_' +
                           self.end_date.date().strftime("%Y%m%d"))
                self.figs_path = os.path.join(self.save_path, '{}/'.format(ext_shr))

                if external_logger == None:
                    createLog(self)
                else:
                    self._logger = external_logger

            raise Exception(log[-1])

        for l in log:
            self.tmp_log.append(l)

        if out['dates'] == []:
            raise Exception('Supplied [run] directory, start_date, and end_date '
                            'give no valid snow files')

        self.outputs = out
        self.run_dirs = dirs
        self.all_dirs = all_dirs
        self.rundirs_dict = rdict
        self.all_dirs_flt = deepcopy(all_dirs)

        if self.start_date is not None and self.end_date is None:
            self.end_date = self.outputs['dates'][-1]
            self.tmp_log.append(' Config options [run] end_date '
                                'not specified, assigning '
                                '{} and {}'.format(self.start_date, self.end_date))

            self.ixs = 0
            self.ixe = len(self.outputs['dates']) - 1

        # Otherwise, get closest dates and make indices
        else:
            self.start_date = self.outputs['dates'][0]
            self.end_date = self.outputs['dates'][-1]
            self.ixs = 0
            self.ixe = len(self.outputs['dates']) - 1

        if ((self.start_date.date() < self.outputs['dates'][0].date())
                or (self.end_date.date() > self.outputs['dates'][-1].date())):
            raise Exception('ERROR! Config option [run] start_date or end_date '
                            'outside of date range found in [run] directory')

        # Since model outputs at 23:00, step the figure and report dates to
        # show 00:00 the next day (unless start of water year)
        if self.start_date == datetime(self.wy - 1, 10, 1, 23, 0, 0):
            self.report_start = self.start_date

        else:
            self.report_start = self.start_date + timedelta(hours=1)

        # Copy the config file where figs will be saved
        # use directory if only plotting figures from database and don't
        # have start_date, end_date
        extf = os.path.splitext(os.path.split(self.config_file)[1])
        ext_shr = (self.directory +
                   '_' +
                   self.start_date.date().strftime("%Y%m%d") +
                   '_' +
                   self.end_date.date().strftime("%Y%m%d"))
        self.figs_path = os.path.join(self.save_path, '{}/'.format(ext_shr))

        # get forecast outputs
        if self.forecast_flag:
            results = outputs(self.for_run_dirs, self.wy, self.properties,
                              None, None, None, self.loglevel)

            self.for_outputs = results['outputs']
            self.for_run_dirs = results['run_dirs']
            self.for_rundirs_dict = results['rdict']
            self.for_ixs = 0
            self.for_ixe = len(self.for_outputs['swe_z']) - 1

        # Get outputs for flights
        if self.flt_flag:

            file = self.update_file
            p = nc.Dataset(file, 'r')

            if self.update_numbers is None:
                times = p.variables['time'][:]
            else:
                if sum([x > len(p.variables['time']) for x in self.update_numbers]) > 0:
                    self.tmp_log.append(' Value in [plots] update_numbers out of '
                                        'range, max is {}, flight update figs '
                                        'being set to False'.format(len(p.variables['time'])))
                    times = []
                    self.flt_flag = False

                else:
                    times = p.variables['time'][self.update_numbers]

            p.close()

            flight_dates = []
            pre_flight_dates = []

            for time in times:
                wydate = calculate_date_from_wyhr(int(time), self.wy)
                pre_wydate = calculate_date_from_wyhr(int(time - 24), self.wy)
                flight_dates = np.append(flight_dates, wydate)
                pre_flight_dates = np.append(pre_flight_dates, pre_wydate)

            if self.loglevel == 'DEBUG' and self.log_to_file is not True:
                print('Reading files in {} for flight updates'
                      '...'.format(self.run_dirs[0].split('runs')[0]))

            results = outputs(self.all_dirs_flt, self.wy, self.properties,
                              None, None, flight_dates, self.loglevel)

            self.flight_outputs = results['outputs']
            self.run_dirs_flt = results['run_dirs']
            self.flt_rundirs_dict = results['rdict']
            self.flight_diff_dates = results['outputs']['dates']
            self.pre_flight_outputs = results['outputs']

            results = outputs(self.all_dirs_flt, self.wy, self.properties,
                              None, None, pre_flight_dates, self.loglevel)

            self.pre_flight_outputs = results['outputs']

            # If there are no flights in the period, set to false for the flight
            # difference figure and report
            if not self.run_dirs_flt:
                self.flt_flag = False
                self.tmp_log.append(' Config option [plots] update_file was '
                                    'supplied, but no snow.nc files were found in '
                                    '[run] directory that fit the date range, no '
                                    'flight difference figure will be made')

        self.report_date = self.end_date + timedelta(hours=1)
        parts = self.report_name.split('.')
        self.report_name = (parts[0] + self.report_date.date().strftime("%Y%m%d") +
                            '.' + parts[1])

        if not os.path.exists(self.figs_path):
            os.makedirs(self.figs_path)

        config_copy = '{}{}{}'.format(self.figs_path, ext_shr, extf[1])
        generate_config(self.ucfg, config_copy)

        if external_logger == None:
            createLog(self)
        else:
            self._logger = external_logger

        if self.inputs_basins is None:
            self.inputs_basins = [self.plotorder[0]]


def createLog(self):
    '''
    Create log file and print out saved logging statements.
    '''

    level_styles = {'info': {'color': 'white'},
                    'notice': {'color': 'magenta'},
                    'verbose': {'color': 'blue'},
                    'success': {'color': 'green', 'bold': True},
                    'spam': {'color': 'green', 'faint': True},
                    'critical': {'color': 'red', 'bold': True},
                    'error': {'color': 'red'},
                    'debug': {'color': 'green'},
                    'warning': {'color': 'yellow'}}

    field_styles = {'hostname': {'color': 'magenta'},
                    'programname': {'color': 'cyan'},
                    'name': {'color': 'white'},
                    'levelname': {'color': 'white', 'bold': True},
                    'asctime': {'color': 'green'}}

    # start logging
    loglevel = self.loglevel
    numeric_level = getattr(logging, loglevel, None)

    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % loglevel)

    # setup the logging
    logfile = None
    if self.log_to_file:
        logfile = os.path.join(self.figs_path, 'log_snowav.out')
        # let user know
        print('Logging to file: {}'.format(logfile))

    fmt = '%(levelname)s:%(module)s:%(message)s'

    if logfile is not None:

        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        logging.basicConfig(filename=logfile,
                            filemode='w',
                            level=numeric_level,
                            format=fmt)
    else:
        logging.basicConfig(level=numeric_level)
        coloredlogs.install(level=numeric_level,
                            fmt=fmt,
                            level_styles=level_styles,
                            field_styles=field_styles)

    self._loglevel = numeric_level

    self._logger = logging.getLogger(__name__)

    if len(self.tmp_log) > 0:
        for l in self.tmp_log:
            self._logger.info(l)
    if len(self.tmp_warn) > 0:
        for l in self.tmp_warn:
            self._logger.warning(l)
    if len(self.tmp_err) > 0:
        for l in self.tmp_err:
            self._logger.error(l)



from snowav.utils.OutputReader import iSnobalReader
import numpy as np
import os
import copy
import pandas as pd
from datetime import datetime, timedelta
import logging
import coloredlogs
import netCDF4 as nc
from snowav import database
import warnings
from collections import OrderedDict
from snowav.utils.utilities import masks
import snowav.utils.get_topo_stats as ts
from snowav.utils.wyhr import handle_year_stradling, calculate_date_from_wyhr


def process(nc_path, topo_path, value, log):
    '''
    Process single day snow.nc files from command line call.

    Currently only processing swe_z and swe_vol values.

    Args
    ------
    nc_path : str
    topo_path : str
    value : str
    log : object

    '''

    log.info(' Using topo {}...'.format(topo_path))

    out = masks(topo_path, False)
    dem = out['dem'] * 3.28
    bmask = out['masks']
    plotorder = out['plotorder']
    labels = out['labels']
    depth_factor = 0.03937
    snowbands = [0, 1, 2]
    embands = [6, 7, 8, 9]
    cclimit = -5 * 1000 * 1000
    edges = np.arange(4000, 15000, 1000)
    def_edges = np.arange(3000, 14000, 1000)
    ixd = np.digitize(dem, edges)

    if type(nc_path) != list:
        nc_path = [nc_path]
    else:
        nc_path = nc_path

    topo = ts.get_topo_stats(topo_path)
    pixel = int(topo['dv'])
    conversion_factor = ((pixel ** 2) * 0.000000810713194 * 0.001)

    ncf = nc.Dataset(nc_path[0])
    t = nc.num2date(ncf.variables['time'][0], ncf.variables['time'].units)
    ncf.close()
    wy = handle_year_stradling(t) + 1

    vars = OrderedDict([('coldcont', 'cold content'),
                        ('swe_z', 'snow water equivalent depth'),
                        ('swe_vol', 'snow water equivalent volume'),
                        ('swe_avail', 'snow water equivalent available for melt'),
                        ('swe_unavail', 'snow water equivalent unavailable for melt')])

    results = {}
    results['df'] = {}
    outputs = {'swi_z': [], 'evap_z': [], 'snowmelt': [], 'swe_z': [], 'depth': [],
               'dates': [], 'time': [], 'density': [], 'coldcont': [], 'path': []}

    rundirs_dict = {}
    for ncp in nc_path:
        log.info(' Loading {}...'.format(ncp))

        path = ncp.split('snow.nc')[0]
        output = iSnobalReader(path, snowbands=snowbands, embands=embands,
                               wy=wy)

        for n in range(0, len(output.em_data[8])):
            outputs['swi_z'].append(output.em_data[8][n, :, :])
            outputs['snowmelt'].append(output.em_data[7][n, :, :])
            outputs['evap_z'].append(output.em_data[6][n, :, :])
            outputs['coldcont'].append(output.em_data[9][n, :, :])
            outputs['swe_z'].append(output.snow_data[2][n, :, :])
            outputs['depth'].append(output.snow_data[0][n, :, :])
            outputs['density'].append(output.snow_data[1][n, :, :])

        outputs['dates'] = np.append(outputs['dates'], output.dates)
        outputs['time'] = np.append(outputs['time'], output.time)
        outputs['path'] = ncp

    dz = pd.DataFrame(0, index=def_edges, columns=bmask.keys())

    if len(outputs['dates']) > 1:
        if outputs['dates'][0] == outputs['dates'][1]:
            outputs['dates'][1] = outputs['dates'][1] + timedelta(hours=1 / 60)

    log.info(' Generating results for {}...'.format(value))

    for iters, out_date in enumerate(outputs['dates']):

        # Initialize output dataframes for every day
        daily_outputs = {'swe_vol': dz.copy(),
                         'swe_avail': dz.copy(),
                         'swe_unavail': dz.copy(),
                         'swe_z': dz.copy(),
                         'coldcont': dz.copy()}

        # Get a snow-free mask ready
        swe = copy.deepcopy(outputs['swe_z'][iters])
        cold = copy.deepcopy(outputs['coldcont'][iters])

        # Loop over outputs (depths are copied, volumes are calculated)
        for k in vars.keys():

            # Mask by subbasin
            for name in bmask:
                mask = copy.deepcopy(bmask[name]['mask'])
                mask = mask.astype(float)

                mask[mask < 1] = np.nan
                elevbin = ixd * mask

                # If the key is something we've already calculated (i.e., not
                # volume), get the masked subbasin output, otherwise it will be
                # calculated below
                if k in outputs.keys():
                    mask_out = outputs[k][iters] * mask

                # make a subbasin, by elevation, snow-free mask
                swe_mask_sub = copy.deepcopy(swe) * mask
                cold_sub = cold * mask

                # Now mask the subbasins by elevation band
                for n in np.arange(0, len(def_edges)):
                    ind = elevbin == n
                    b = def_edges[n]

                    # index for pixels with SWE, in subbasin, in elevation band,
                    # needed for mean values
                    ix = swe_mask_sub[ind] > 0

                    # Assign values
                    if k == 'swe_z':
                        mask_out = outputs['swe_z'][iters] * mask
                        ccb = cold_sub[ind]
                        cind = ccb > cclimit

                        # masked out by pixels with snow -> [ix]
                        swe_bin = mask_out[ind]

                        if swe_bin.size:
                            r = np.nansum(swe_bin[cind])
                            ru = np.nansum(swe_bin[~cind])

                            daily_outputs['swe_unavail'].loc[b, name] = ru * conversion_factor
                            daily_outputs['swe_avail'].loc[b, name] = r * conversion_factor
                            daily_outputs['swe_z'].loc[b, name] = np.nanmean(swe_bin) * depth_factor
                            daily_outputs['swe_vol'].loc[b, name] = np.nansum(swe_bin) * conversion_factor

                        else:
                            daily_outputs['swe_unavail'].loc[b, name] = np.nan
                            daily_outputs['swe_avail'].loc[b, name] = np.nan
                            daily_outputs['swe_z'].loc[b, name] = np.nan
                            daily_outputs['swe_vol'].loc[b, name] = np.nan

                # Add basin total mean/volume field once all elevations are done
                isub = swe_mask_sub > 0

                if k == 'swe_z':
                    # calc swe_z derived values
                    out = outputs[k][iters] * mask
                    total = np.nanmean(out) * depth_factor

                    s = np.nansum(daily_outputs['swe_vol'][name].values)
                    s1 = np.nansum(daily_outputs['swe_avail'][name].values)
                    s2 = np.nansum(daily_outputs['swe_unavail'][name].values)

            if k == 'swe_vol':
                df = daily_outputs[k].copy()
                df = df.round(decimals=3)

        results['df'][nc_path[iters]] = df.fillna(0)

    results['outputs'] = outputs
    results['masks'] = bmask
    results['plotorder'] = plotorder
    results['edges'] = def_edges
    results['labels'] = labels

    return results
